from django.contrib import messages
from django.contrib.admin import ModelAdmin, TabularInline
from django.db.models import Count
from django.forms import BaseInlineFormSet
from django.shortcuts import get_object_or_404
from django.utils.html import format_html

from django_ledger.models import (
    LedgerModel, JournalEntryModel, EntityModel,
    LedgerModelValidationError
)


class JournalEntryModelInLineFormSet(BaseInlineFormSet):

    def __init__(self, *args, **kwargs):
        self.ledger_model: LedgerModel = kwargs['instance']
        self.entity_model = self.ledger_model.entity
        super().__init__(*args, **kwargs)

    def add_fields(self, form, index):
        super().add_fields(form=form, index=index)
        form.fields['entity_unit'].queryset = self.entity_model.entityunitmodel_set.all()


class JournalEntryModelInLine(TabularInline):
    extra = 0
    fields = [
        'timestamp',
        'description',
        'entity_unit',
        'posted',
        'locked',
        'activity',
        'origin',
        'txs_count',
        'view_txs_link',
        'edit_txs_link'
    ]
    readonly_fields = [
        'posted',
        'locked',
        'origin',
        'activity',
        'txs_count',
        'view_txs_link',
        'edit_txs_link'
    ]
    model = JournalEntryModel
    formset = JournalEntryModelInLineFormSet

    def get_queryset(self, request):
        qs = self.model.objects.for_user(request.user)
        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(*ordering)
        return qs.annotate(
            txs_count=Count('transactionmodel')
        ).select_related(
            'ledger',
            'ledger__entity'
        )

    def txs_count(self, obj):
        return obj.txs_count

    txs_count.short_description = 'Transactions'

    def has_change_permission(self, request, obj=None):
        if obj:
            return all([
                not obj.is_locked(),
                super().has_change_permission(request, obj)
            ])
        return super().has_change_permission(request, obj)

    def has_delete_permission(self, request, obj=None):
        if obj:
            return all([
                obj.can_delete(),
                super().has_delete_permission(request, obj)
            ])
        return super().has_delete_permission(request, obj)

    def view_txs_link(self, obj: JournalEntryModel):
        detail_url = obj.get_detail_url()
        return format_html(
            format_string='<a class="viewlink" target="_blank" href="{url}">View Txs</a>',
            url=detail_url
        )

    def edit_txs_link(self, obj: JournalEntryModel):
        detail_url = obj.get_detail_txs_url()
        next_url = None
        return format_html(
            format_string='<a class="changelink" target="_blank" href="{url}">Edit Txs</a>',
            url=detail_url
        )


class LedgerModelAdmin(ModelAdmin):
    readonly_fields = [
        'entity',
        'posted',
        'locked'
    ]
    list_filter = [
        'posted',
        'locked',
        'entity__name'
    ]
    list_display = [
        'name',
        'ledger_xid',
        'is_posted',
        'is_locked',
        'is_extended',
        'journal_entry_count',
        'earliest_journal_entry',
        'created',
        'updated'
    ]
    actions = [
        'post',
        'unpost',
        'lock',
        'unlock'
    ]
    inlines = [
        JournalEntryModelInLine
    ]

    def get_queryset(self, request):
        qs = LedgerModel.objects.for_user(user_model=request.user)
        ordering = self.get_ordering(request)
        if ordering:
            qs = qs.order_by(*ordering)
        return qs.select_related('entity')

    def get_inlines(self, request, obj):
        if obj is None:
            return []
        return super().get_inlines(request, obj)

    def get_fieldsets(self, request, obj=None):
        if obj is None:
            return [
                ('', {
                    'fields': [
                        'name',
                        'hidden',
                        'additional_info'
                    ]
                })
            ]
        return [
            ('', {
                'fields': [
                    'entity',
                    'name',
                    'hidden',
                    'additional_info'
                ]
            })
        ]

    def get_entity_model(self, request) -> EntityModel:
        entity_slug = request.GET.get('entity_slug')
        entity_model_qs = EntityModel.objects.for_user(user_model=request.user)
        entity_model = get_object_or_404(entity_model_qs, slug__exact=entity_slug)
        return entity_model

    def has_add_permission(self, request):
        if request.GET.get('entity_slug') is not None:
            return True
        return False

    def add_view(self, request, form_url="", extra_context=None):
        entity_model = self.get_entity_model(request)
        extra_context = {
            'entity_model': entity_model,
            'title': f'Add Ledger: {entity_model.name}'
        }
        return super().add_view(request, form_url="", extra_context=extra_context)

    def is_locked(self, obj):
        return obj.is_locked()

    is_locked.boolean = True

    def is_posted(self, obj):
        return obj.is_posted()

    is_posted.boolean = True

    def is_extended(self, obj):
        return obj.has_wrapped_model()

    is_extended.boolean = True

    # ACTIONS....
    def post(self, request, queryset):
        for obj in queryset:
            try:
                obj.post(raise_exception=True, commit=False)
            except LedgerModelValidationError as e:
                messages.error(
                    request=request,
                    message=e.message
                )
        queryset.bulk_update(
            objs=queryset,
            fields=[
                'posted',
                'updated'
            ])

    def lock(self, request, queryset):
        for obj in queryset:
            try:
                obj.lock(raise_exception=True, commit=False)
            except LedgerModelValidationError as e:
                messages.error(
                    request=request,
                    message=e.message
                )

        queryset.bulk_update(
            objs=queryset,
            fields=[
                'locked',
                'updated'
            ])

    def unpost(self, request, queryset):
        for obj in queryset:
            try:
                obj.unpost(raise_exception=True, commit=False)
            except LedgerModelValidationError as e:
                messages.error(
                    request=request,
                    message=e.message
                )
        queryset.bulk_update(
            objs=queryset,
            fields=[
                'posted',
                'updated'
            ])

    def unlock(self, request, queryset):
        for obj in queryset:
            try:
                obj.unlock(raise_exception=True, commit=False)
            except LedgerModelValidationError as e:
                messages.error(
                    request=request,
                    message=e.message
                )

        queryset.bulk_update(
            objs=queryset,
            fields=[
                'locked',
                'updated'
            ])

    def journal_entry_count(self, obj):
        return obj.journal_entries__count

    def earliest_journal_entry(self, obj):
        return obj.earliest_timestamp

    def save_model(self, request, obj: LedgerModel, form, change):
        if not change:
            entity_model = self.get_entity_model(request)
            obj.entity = entity_model
        return super().save_model(
            request=request,
            obj=obj,
            form=form,
            change=change
        )

    class Meta:
        model = LedgerModel



from django.urls import path

from django_ledger import views

urlpatterns = [

    path('<slug:entity_slug>/<slug:coa_slug>/create/',
         views.AccountModelCreateView.as_view(),
         name='account-create'),
    path('<slug:entity_slug>/<slug:coa_slug>/list/',
         views.AccountModelListView.as_view(),
         name='account-list'),
    path('<slug:entity_slug>/<slug:coa_slug>/list/active/',
         views.AccountModelListView.as_view(active_only=True),
         name='account-list-active'),

    # Account Transaction Detail....
    path('<slug:entity_slug>/<slug:coa_slug>/update/<uuid:account_pk>/',
         views.AccountModelUpdateView.as_view(),
         name='account-update'),
    path('<slug:entity_slug>/<slug:coa_slug>/detail/<uuid:account_pk>/',
         views.AccountModelDetailView.as_view(),
         name='account-detail'),
    path('<slug:entity_slug>/<slug:coa_slug>/detail/<uuid:account_pk>/year/<int:year>/',
         views.AccountModelYearDetailView.as_view(),
         name='account-detail-year'),
    path('<slug:entity_slug>/<slug:coa_slug>/detail/<uuid:account_pk>/quarter/<int:year>/<int:quarter>/',
         views.AccountModelQuarterDetailView.as_view(),
         name='account-detail-quarter'),
    path('<slug:entity_slug>/<slug:coa_slug>/detail/<uuid:account_pk>/month/<int:year>/<int:month>/',
         views.AccountModelMonthDetailView.as_view(),
         name='account-detail-month'),
    path('<slug:entity_slug>/<slug:coa_slug>/detail/<uuid:account_pk>/date/<int:year>/<int:month>/<int:day>/',
         views.AccountModelDateDetailView.as_view(),
         name='account-detail-date'),

    # Account Actions...
    path('<slug:entity_slug>/<slug:coa_slug>/action/<uuid:account_pk>/activate/',
         views.BaseAccountModelActionView.as_view(action_name='activate'),
         name='account-action-activate'),
    path('<slug:entity_slug>/<slug:coa_slug>/action/<uuid:account_pk>/deactivate/',
         views.BaseAccountModelActionView.as_view(action_name='deactivate'),
         name='account-action-deactivate'),
    path('<slug:entity_slug>/<slug:coa_slug>/action/<uuid:account_pk>/lock/',
         views.BaseAccountModelActionView.as_view(action_name='lock'),
         name='account-action-lock'),
    path('<slug:entity_slug>/<slug:coa_slug>/action/<uuid:account_pk>/unlock/',
         views.BaseAccountModelActionView.as_view(action_name='unlock'),
         name='account-action-unlock')
]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from django.forms import ModelForm, TextInput, EmailInput, NumberInput
from django.utils.translation import gettext_lazy as _

from django_ledger.forms.utils import validate_cszc
from django_ledger.models.customer import CustomerModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class CustomerModelForm(ModelForm):

    def clean(self):
        validate_cszc(self.cleaned_data)

    class Meta:
        model = CustomerModel
        fields = [
            'customer_name',
            'customer_code',
            'address_1',
            'address_2',
            'city',
            'state',
            'zip_code',
            'country',
            'phone',
            'email',
            'website',
            'sales_tax_rate',
            'active',
            'hidden'
        ]
        help_texts = {
            'sales_tax_rate': _('Example: 3.50% should be entered as 0.035')
        }
        widgets = {
            'customer_name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'customer_code': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': 'Ex: CDR-78987-2025...'
            }),
            'address_1': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'address_2': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'city': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'state': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'zip_code': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'country': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
            }),
            'phone': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
            }),
            'email': EmailInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'website': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'sales_tax_rate': NumberInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'min': 0.000,
                'max': 1.000,
                'step': 0.001
            })
        }



from django.urls import path
from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/list/', views.VendorModelListView.as_view(), name='vendor-list'),
    path('<slug:entity_slug>/create/', views.VendorModelCreateView.as_view(), name='vendor-create'),
    path('<slug:entity_slug>/update/<uuid:vendor_pk>/', views.VendorModelUpdateView.as_view(), name='vendor-update'),
]


from django.forms import ModelForm, TextInput, Select, ValidationError
from django.utils.translation import gettext_lazy as _

from django_ledger.io.roles import ASSET_CA_CASH, LIABILITY_CL_ACC_PAYABLE, LIABILITY_LTL_MORTGAGE_PAYABLE
from django_ledger.models import BankAccountModel
from django_ledger.models.accounts import AccountModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class BankAccountCreateForm(ModelForm):

    def __init__(self, *args, entity_slug, user_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model

        # todo: only the accounts that do not hava an associated bank account should be available to pick from...
        account_qs = AccountModel.objects.for_entity(
            user_model=self.USER_MODEL,
            entity_model=self.ENTITY_SLUG
        ).available().filter(
            role__in=[
                ASSET_CA_CASH,
                LIABILITY_CL_ACC_PAYABLE,
                LIABILITY_LTL_MORTGAGE_PAYABLE
            ])
        self.fields['account_model'].queryset = account_qs

    def clean(self):
        account_model = self.cleaned_data['account_model']
        routing_number = self.cleaned_data['routing_number']
        account_number = self.cleaned_data['account_number']

        if not account_model:
            raise ValidationError('Must select a bank account.')

        # catching unique database constraint...
        if BankAccountModel.objects.filter(
                account_model=account_model,
                routing_number__exact=routing_number,
                account_number__exact=account_number
        ).exists():
            raise ValidationError('Duplicate bank account model.')

    class Meta:
        model = BankAccountModel
        fields = [
            'name',
            'account_type',
            'account_number',
            'routing_number',
            'aba_number',
            'swift_number',
            'account_model',
            'active'
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Enter account name...')
            }),
            'account_number': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Enter account number...')
            }),
            'routing_number': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Enter routing number...')
            }),
            'aba_number': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Enter ABA number...')
            }),
            'swift_number': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Enter SWIFT number...')
            }),
            'account_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'account_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            })
        }
        labels = {
            'name': _('Account Name'),
            'account_number': _('Account Number'),
            'account_type': _('Account Type'),
            'account_model': _('CoA Account'),
            'aba_number': _('ABA Number'),
            'routing_number': _('Routing Number'),
            'active': _('Make Active'),
        }


class BankAccountUpdateForm(BankAccountCreateForm):
    class Meta:
        model = BankAccountModel
        fields = [
            'name',
            'account_type',
            'account_model',
            'active'
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('Enter account name...')
            }),
            'account_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'account_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            })
        }

    def clean(self):
        cash_account = self.cleaned_data['account_model']

        if not cash_account:
            raise ValidationError('Must select a bank account.')

        # catching unique database constraint...
        if 'cash_account' in self.changed_data:
            if BankAccountModel.objects.filter(
                    cash_account=cash_account,
                    routing_number__exact=self.instance.routing_number,
                    account_number__exact=self.instance.account_number
            ).exists():
                raise ValidationError('Duplicate bank account model.')



from django.urls import path
from django_ledger.views.feedback import BugReportView, RequestNewFeatureView

urlpatterns = [
    path('bug-report/', BugReportView.as_view(), name='bug-report'),
    path('new-feature/', RequestNewFeatureView.as_view(), name='new-feature')
]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import (
    DetailView, UpdateView, CreateView,
    RedirectView, DeleteView, ArchiveIndexView, YearArchiveView, MonthArchiveView
)
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.ledger import LedgerModelCreateForm, LedgerModelUpdateForm
from django_ledger.io.io_core import get_localdate
from django_ledger.models import EntityModel
from django_ledger.models.ledger import LedgerModel
from django_ledger.views.mixins import (
    YearlyReportMixIn, QuarterlyReportMixIn,
    MonthlyReportMixIn, DjangoLedgerSecurityMixIn, DateReportMixIn, BaseDateNavigationUrlMixIn,
    EntityUnitMixIn, PDFReportMixIn
)


class LedgerModelModelBaseView(DjangoLedgerSecurityMixIn):
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            entity_model: EntityModel = self.get_authorized_entity_instance()
            self.queryset = entity_model.get_ledgers()
        return self.queryset


class LedgerModelListView(LedgerModelModelBaseView, ArchiveIndexView):
    allow_empty = True
    context_object_name = 'ledger_list'
    template_name = 'django_ledger/ledger/ledger_list.html'
    PAGE_TITLE = _('Entity Ledgers')
    paginate_by = 30
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }
    date_field = 'created'
    ordering = '-created'


    show_all = False
    show_current = False
    show_visible = False

    def get_queryset(self):
        qs = super().get_queryset()
        qs = qs.select_related('billmodel', 'invoicemodel')
        qs = qs.order_by('-created')
        if self.show_all:
            return qs
        if self.show_current:
            qs = qs.current()
        if self.show_visible:
            qs = qs.visible()
        return qs

    def get_context_data(self, *, object_list=None, **kwargs):
        ctx = super().get_context_data(object_list=object_list, **kwargs)
        ctx['page_title'] = self.PAGE_TITLE
        ctx['header_title'] = self.PAGE_TITLE
        if self.show_all:
            ctx['header_subtitle'] = 'All Ledger Models'
            ctx['header_subtitle_icon'] = 'emojione-monotone:ledger'
        if self.show_visible:
            ctx['header_subtitle'] = 'Visible Ledger Models'
            ctx['header_subtitle_icon'] = 'mingcute:user-visible-line'
        if self.show_current:
            ctx['header_subtitle'] = 'Current Ledger Models'
            ctx['header_subtitle_icon'] = 'basil:current-location-outline'
        return ctx


class LedgerModelYearListView(YearArchiveView, LedgerModelListView):
    make_object_list = True


class LedgerModelMonthListView(MonthArchiveView, LedgerModelListView):
    make_object_list = True


class LedgerModelCreateView(LedgerModelModelBaseView, CreateView):
    template_name = 'django_ledger/ledger/ledger_create.html'
    PAGE_TITLE = _('Create Ledger')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }

    def get_form(self, form_class=None):
        return LedgerModelCreateForm(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            **self.get_form_kwargs()
        )

    def form_valid(self, form):
        instance = form.save(commit=False)
        instance.entity = self.AUTHORIZED_ENTITY_MODEL
        return super().form_valid(form)

    def get_success_url(self):
        return reverse('django_ledger:ledger-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })


class LedgerModelDetailView(LedgerModelModelBaseView, RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        return reverse('django_ledger:je-list',
                       kwargs={
                           'ledger_pk': self.kwargs['ledger_pk'],
                           'entity_slug': self.kwargs['entity_slug']
                       })


class LedgerModelUpdateView(LedgerModelModelBaseView, UpdateView):
    context_object_name = 'ledger'
    pk_url_kwarg = 'ledger_pk'
    template_name = 'django_ledger/ledger/ledger_update.html'

    def get_form(self, form_class=None):
        return LedgerModelUpdateForm(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            **self.get_form_kwargs()
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = _('Update Ledger: ') + self.object.name
        context['header_title'] = context['page_title']
        return context

    def get_success_url(self):
        return reverse('django_ledger:ledger-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })


class LedgerModelDeleteView(LedgerModelModelBaseView, DeleteView):
    template_name = 'django_ledger/ledger/ledger_delete.html'
    pk_url_kwarg = 'ledger_pk'
    context_object_name = 'ledger_model'

    def get_success_url(self):
        return reverse(viewname='django_ledger:ledger-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })


# ACTIONS....

class LedgerModelModelActionView(LedgerModelModelBaseView,
                                 RedirectView,
                                 SingleObjectMixin):
    http_method_names = ['get']
    pk_url_kwarg = 'ledger_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, *args, **kwargs):
        return reverse('django_ledger:ledger-list',
                       kwargs={
                           'entity_slug': kwargs['entity_slug']
                       })

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(LedgerModelModelActionView, self).get(request, *args, **kwargs)
        ledger_model: LedgerModel = self.get_object()

        try:
            getattr(ledger_model, self.action_name)(commit=self.commit, **kwargs)
        except ValidationError as e:
            messages.add_message(request,
                                 message=e.message,
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
        return response


# Ledger Balance Sheet Views...
class BaseLedgerModelBalanceSheetView(LedgerModelModelBaseView, RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        year = get_localdate().year
        return reverse('django_ledger:ledger-bs-year', kwargs={
            'entity_slug': self.kwargs['entity_slug'],
            'ledger_pk': self.kwargs['ledger_pk'],
            'year': year
        })


class FiscalYearLedgerModelBalanceSheetView(LedgerModelModelBaseView,
                                            BaseDateNavigationUrlMixIn,
                                            EntityUnitMixIn,
                                            YearlyReportMixIn,
                                            PDFReportMixIn,
                                            DetailView):
    context_object_name = 'ledger'
    pk_url_kwarg = 'ledger_pk'
    template_name = 'django_ledger/financial_statements/balance_sheet.html'
    pdf_report_type = 'BS'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = _('Ledger Balance Sheet: ') + self.object.name
        context['header_title'] = context['page_title']
        return context


class QuarterlyLedgerModelBalanceSheetView(FiscalYearLedgerModelBalanceSheetView, QuarterlyReportMixIn):
    """
    Quarter Balance Sheet View.
    """


class MonthlyLedgerModelBalanceSheetView(FiscalYearLedgerModelBalanceSheetView, MonthlyReportMixIn):
    """
    Monthly Balance Sheet View.
    """


class DateLedgerModelBalanceSheetView(FiscalYearLedgerModelBalanceSheetView, DateReportMixIn):
    """
    Date Balance Sheet View.
    """


# Ledger Income Statement Views...
class BaseLedgerIncomeStatementView(LedgerModelModelBaseView, RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        year = get_localdate().year
        return reverse('django_ledger:ledger-ic-year',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug'],
                           'ledger_pk': self.kwargs['ledger_pk'],
                           'year': year
                       })


class FiscalYearLedgerIncomeStatementView(LedgerModelModelBaseView,
                                          BaseDateNavigationUrlMixIn,
                                          EntityUnitMixIn,
                                          YearlyReportMixIn,
                                          PDFReportMixIn,
                                          DetailView):
    context_object_name = 'ledger'
    pk_url_kwarg = 'ledger_pk'
    template_name = 'django_ledger/financial_statements/income_statement.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = _('Ledger Income Statement: ') + self.object.name
        context['header_title'] = context['page_title']
        return context


class QuarterlyLedgerIncomeStatementView(FiscalYearLedgerIncomeStatementView, QuarterlyReportMixIn):
    """
    Quarterly Income Statement Quarter Report.
    """


class MonthlyLedgerIncomeStatementView(FiscalYearLedgerIncomeStatementView, MonthlyReportMixIn):
    """
    Monthly Income Statement Monthly Report.
    """


class DateLedgerIncomeStatementView(FiscalYearLedgerIncomeStatementView, DateReportMixIn):
    """
    Date Income Statement Monthly Report.
    """


# CASH FLOW STATEMENT ----
class BaseLedgerModelCashFlowStatementRedirectView(LedgerModelModelBaseView, RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        year = get_localdate().year
        return reverse('django_ledger:ledger-cf-year',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug'],
                           'ledger_pk': self.kwargs['ledger_pk'],
                           'year': year
                       })


class FiscalYearLedgerModelCashFlowStatementView(LedgerModelModelBaseView,
                                                 BaseDateNavigationUrlMixIn,
                                                 EntityUnitMixIn,
                                                 YearlyReportMixIn,
                                                 PDFReportMixIn,
                                                 DetailView):
    """
    Fiscal Year Cash Flow Statement View.
    """

    context_object_name = 'ledger'
    pk_url_kwarg = 'ledger_pk'
    template_name = 'django_ledger/financial_statements/cash_flow.html'
    pdf_report_type = 'CFS'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = _('Ledger Cash Flow Statement: ') + self.object.name
        context['header_title'] = context['page_title']
        return context


class QuarterlyLedgerModelCashFlowStatementView(FiscalYearLedgerModelCashFlowStatementView, QuarterlyReportMixIn):
    """
    Quarter Cash Flow Statement View.
    """


class MonthlyLedgerModelCashFlowStatementView(FiscalYearLedgerModelCashFlowStatementView, MonthlyReportMixIn):
    """
    Monthly Cash Flow Statement View.
    """


class DateLedgerModelCashFlowStatementView(FiscalYearLedgerModelCashFlowStatementView, DateReportMixIn):
    """
    Date Cash Flow Statement View.
    """



from django.urls import reverse

from dev_env.settings import LOGIN_URL
from django_ledger.tests.base import DjangoLedgerBaseTest


class AuthTest(DjangoLedgerBaseTest):
    TEST_URLS = [
        reverse('django_ledger:home'),
        reverse('django_ledger:entity-dashboard',
                kwargs={
                    'entity_slug': 'fake-123'
                })
    ]

    def test_user_is_redirected(self):
        for test_url in self.TEST_URLS:
            response = self.client.get(test_url, follow=False)
            login_url = LOGIN_URL + f'?next={test_url}'
            self.assertRedirects(response, expected_url=login_url)

    def test_login_form(self):
        response = self.client.get(LOGIN_URL)
        self.assertContains(response, text='id="djl-el-login-form"', status_code=200, count=1)
        self.assertContains(response, text='id="djl-el-login-form-username-field"', status_code=200, count=1)
        self.assertContains(response, text='id="djl-el-login-form-password-field"', status_code=200, count=1)

    def test_user_with_wrong_credentials(self):
        response = self.client.post(LOGIN_URL, data={
            'username': self.USERNAME,
            'password': self.PASSWORD + '1',
        }, follow=False)
        self.assertContains(response, text='Login', status_code=200)
        self.assertFormError(field=None,
                             form=response.context_data['form'],
                             errors=[
                                 'Please enter a correct username and password. Note that both fields may be case-sensitive.'
                             ])

        response = self.client.post(LOGIN_URL, data={
            'username': self.USERNAME + 'abc',
            'password': self.PASSWORD,
        }, follow=False)
        self.assertContains(response, text='Login', status_code=200)
        self.assertFormError(field=None,
                             form=response.context_data['form'],
                             errors=[
                                 'Please enter a correct username and password. Note that both fields may be case-sensitive.'
                             ])

    def test_user_login(self):
        response = self.client.post(LOGIN_URL, data={
            'username': self.USERNAME,
            'password': self.PASSWORD,
        }, follow=False)
        expeced_url = reverse('django_ledger:home')
        self.assertRedirects(response, expected_url=expeced_url)

    def test_user_can_logout(self):
        home_url = reverse('django_ledger:home')
        login_url = reverse('django_ledger:login')
        logout_url = reverse('django_ledger:logout')
        self.client.login(
            username=self.USERNAME,
            password=self.PASSWORD
        )
        response = self.client.get(home_url)

        # logout button is present...
        self.assertContains(response, text='id="djl-el=logout-button-nav"')

        response = self.client.post(logout_url)
        self.assertRedirects(response, expected_url=login_url)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

SCHEMA_ROLE_ACCOUNT = {
    'type': 'object'
}

SCHEMA_ROLE_BALANCE = {
    'type': 'object'
}

SCHEMA_ROLE_BALANCE_BY_PERIOD = {
    'type': ['object', 'null']
}

SCHEMA_GROUP_ACCOUNT = {
    'type': 'object'
}

SCHEMA_GROUP_BALANCE = {
    'type': 'object'
}

SCHEMA_GROUP_BALANCE_BY_PERIOD = {
    'type': ['object', 'null']
}

SCHEMA_RATIOS = {
    'type': 'object'
}

SCHEMA_ACCOUNTS = {
    'type': 'array'
}

SCHEMA_DIGEST = {
    'type': 'object',
    'properties': {
        'accounts': SCHEMA_ACCOUNTS,
        'role_account': SCHEMA_ROLE_ACCOUNT,
        'role_balance': SCHEMA_ROLE_BALANCE,
        'role_balance_by_period': SCHEMA_ROLE_BALANCE_BY_PERIOD,
        'group_account': SCHEMA_GROUP_ACCOUNT,
        'group_balance': SCHEMA_GROUP_BALANCE,
        'group_balance_by_period': SCHEMA_GROUP_BALANCE_BY_PERIOD,
        'ratios': SCHEMA_RATIOS,
    }
}



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import RedirectView, ListView

from django_ledger.models.entity import EntityModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class RootUrlView(RedirectView):

    def get_redirect_url(self, *args, **kwargs):
        if not self.request.user.is_authenticated:
            return reverse('django_ledger:login')
        return reverse('django_ledger:home')


class DashboardView(DjangoLedgerSecurityMixIn, ListView):
    template_name = 'django_ledger/entity/home.html'
    PAGE_TITLE = _('My Dashboard')
    context_object_name = 'entities'
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
    }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['header_subtitle'] = self.request.user.get_full_name()
        context['header_subtitle_icon'] = 'ei:user'
        return context

    def get_queryset(self):
        return EntityModel.objects.for_user(
            user_model=self.request.user
        ).order_by('-created')



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django.core.mail import send_mail
from django.views.generic import FormView

from django_ledger.forms.feedback import BugReportForm, RequestNewFeatureForm
from django_ledger.settings import DJANGO_LEDGER_FEEDBACK_EMAIL_LIST, DJANGO_LEDGER_FEEDBACK_FROM_EMAIL
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn, SuccessUrlNextMixIn


class BugReportView(DjangoLedgerSecurityMixIn,
                    SuccessUrlNextMixIn,
                    FormView):
    http_method_names = ['post']
    form_class = BugReportForm

    def form_valid(self, form):
        form_data = form.cleaned_data
        message = f'How to reproduce?: {form_data["reproduce"]}\n' + \
                  f'Expectation: {form_data["expectation"]}\n' + \
                  f'Device: {form_data["device"]}\n' + \
                  f'From user: {self.request.user.username}\n' + \
                  f'User email: {self.request.user.email}'
        if DJANGO_LEDGER_FEEDBACK_EMAIL_LIST:
            send_mail(
                subject=f'DJL Bug Report',
                from_email=DJANGO_LEDGER_FEEDBACK_FROM_EMAIL,
                recipient_list=DJANGO_LEDGER_FEEDBACK_EMAIL_LIST,
                fail_silently=True,
                message=message)
        return super().form_valid(form)


class RequestNewFeatureView(DjangoLedgerSecurityMixIn,
                            SuccessUrlNextMixIn,
                            FormView):
    http_method_names = ['post']
    form_class = RequestNewFeatureForm

    def form_valid(self, form):
        form_data = form.cleaned_data
        message = f'Description: {form_data["feature_description"]}\n' + \
                  f'Solution: {form_data["solution"]}\n' + \
                  f'Alternatives: {form_data["alternatives"]}\n' + \
                  f'From user: {self.request.user.username}\n' + \
                  f'User email: {self.request.user.email}'
        if DJANGO_LEDGER_FEEDBACK_EMAIL_LIST:
            send_mail(
                subject=f'DJL New Feature Request',
                from_email=DJANGO_LEDGER_FEEDBACK_FROM_EMAIL,
                recipient_list=DJANGO_LEDGER_FEEDBACK_EMAIL_LIST,
                fail_silently=True,
                message=message)
        return super().form_valid(form)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

SCHEMA_PNL_DATA = {
    'type': 'object',
}

SCHEMA_PNL = {
    'type': 'object',
    'properties': {
        'entity_slug': {
            'type': 'string'
        },
        'entity_name': {
            'type': 'string'
        },
        'pnl_data': SCHEMA_PNL_DATA
    },
    'required': [
        'entity_slug',
        'entity_name',
        'pnl_data'
    ]
}



# Generated by Django 5.2.6 on 2025-09-08 21:32

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('django_ledger', '0024_billmodel_entity_model_invoicemodel_entity_model'),
    ]

    operations = [
        migrations.AlterField(
            model_name='billmodel',
            name='cash_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='bill_cash_account', to='django_ledger.accountmodel',
                                    verbose_name='Cash Account'),
        ),
        migrations.AlterField(
            model_name='billmodel',
            name='prepaid_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='bill_prepaid_account', to='django_ledger.accountmodel',
                                    verbose_name='Prepaid Account'),
        ),
        migrations.AlterField(
            model_name='billmodel',
            name='unearned_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='bill_unearned_account', to='django_ledger.accountmodel',
                                    verbose_name='Unearned Account'),
        ),
        migrations.AlterField(
            model_name='customermodel',
            name='address_1',
            field=models.CharField(blank=True, max_length=70, null=True, verbose_name='Address Line 1'),
        ),
        migrations.AlterField(
            model_name='entitymodel',
            name='address_1',
            field=models.CharField(blank=True, max_length=70, null=True, verbose_name='Address Line 1'),
        ),
        migrations.AlterField(
            model_name='invoicemodel',
            name='cash_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='invoice_cash_account', to='django_ledger.accountmodel',
                                    verbose_name='Cash Account'),
        ),
        migrations.AlterField(
            model_name='invoicemodel',
            name='prepaid_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='invoice_prepaid_account', to='django_ledger.accountmodel',
                                    verbose_name='Prepaid Account'),
        ),
        migrations.AlterField(
            model_name='invoicemodel',
            name='unearned_account',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.RESTRICT,
                                    related_name='invoice_unearned_account', to='django_ledger.accountmodel',
                                    verbose_name='Unearned Account'),
        ),
        migrations.AlterField(
            model_name='vendormodel',
            name='address_1',
            field=models.CharField(blank=True, max_length=70, null=True, verbose_name='Address Line 1'),
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from django.contrib.auth.views import LoginView, LogoutView
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _

from django_ledger.forms.auth import LogInForm


class DjangoLedgerLoginView(LoginView):
    form_class = LogInForm
    template_name = 'django_ledger/auth/login.html'
    extra_context = {
        'page_title': _('Login')
    }

    def get_success_url(self):
        return reverse('django_ledger:home')


class DjangoLedgerLogoutView(LogoutView):
    next_page = reverse_lazy('django_ledger:login')



# Generated by Django 4.1.4 on 2023-01-09 17:12

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='journalentrymodel',
            managers=[
            ],
        ),
        migrations.AlterField(
            model_name='transactionmodel',
            name='amount',
            field=models.DecimalField(decimal_places=2, default=0.0, help_text='Account of the transaction.', max_digits=20, validators=[django.core.validators.MinValueValidator(0)], verbose_name='Amount'),
        ),
    ]



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/list/',
         views.ChartOfAccountModelListView.as_view(),
         name='coa-list'),
    path('<slug:entity_slug>/list/inactive/',
         views.ChartOfAccountModelListView.as_view(inactive=True),
         name='coa-list-inactive'),
    path('<slug:entity_slug>/create/',
         views.ChartOfAccountModelCreateView.as_view(),
         name='coa-create'),
    path('<slug:entity_slug>/detail/<slug:coa_slug>/',
         views.ChartOfAccountModelListView.as_view(),
         name='coa-detail'),
    path('<slug:entity_slug>/update/<slug:coa_slug>/',
         views.ChartOfAccountModelUpdateView.as_view(),
         name='coa-update'),

    # ACTIONS....
    path('<slug:entity_slug>/action/<slug:coa_slug>/mark-as-default/',
         views.CharOfAccountModelActionView.as_view(action_name='mark_as_default'),
         name='coa-action-mark-as-default'),
    path('<slug:entity_slug>/action/<slug:coa_slug>/mark-as-active/',
         views.CharOfAccountModelActionView.as_view(action_name='mark_as_active'),
         name='coa-action-mark-as-active'),
    path('<slug:entity_slug>/action/<slug:coa_slug>/mark-as-inactive/',
         views.CharOfAccountModelActionView.as_view(action_name='mark_as_inactive'),
         name='coa-action-mark-as-inactive'),

]



from django.contrib.auth.forms import AuthenticationForm, UsernameField
from django.forms import TextInput, CharField, PasswordInput
from django.utils.translation import gettext_lazy as _


class LogInForm(AuthenticationForm):
    username = UsernameField(
        widget=TextInput(
            attrs={
                'autofocus': True,
                'class': 'input',
                'id': 'djl-el-login-form-username-field'
            }))
    password = CharField(
        label=_("Password"),
        strip=False,
        widget=PasswordInput(
            attrs={
                'autocomplete': 'current-password',
                'class': 'input',
                'id': 'djl-el-login-form-password-field'
            }),
    )



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.
"""
import warnings
from datetime import datetime, time, date
from decimal import Decimal
from itertools import groupby, chain
from typing import Optional
from uuid import uuid4, UUID

from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models
from django.db.models import Q, Manager, QuerySet, Count
from django.db.models.signals import pre_save
from django.urls import reverse
from django.utils.timezone import make_aware
from django.utils.translation import gettext_lazy as _

from django_ledger.models import EntityModel, deprecated_entity_slug_behavior, lazy_loader
from django_ledger.models.journal_entry import JournalEntryModel
from django_ledger.models.ledger import LedgerModel
from django_ledger.models.mixins import CreateUpdateMixIn, MarkdownNotesMixIn
from django_ledger.models.transactions import TransactionModel
from django_ledger.settings import DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR


class ClosingEntryValidationError(ValidationError):
    pass


class ClosingEntryModelQuerySet(QuerySet):

    def for_user(self, user_model):
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(entity_model__admin=user_model) |
            Q(entity_model__managers__in=[user_model])
        )

    def posted(self) -> 'ClosingEntryModelQuerySet':
        return self.filter(posted=True)

    def not_posted(self) -> 'ClosingEntryModelQuerySet':
        return self.filter(posted=False)


class ClosingEntryModelManager(Manager):

    def get_queryset(self) -> ClosingEntryModelQuerySet:
        qs = ClosingEntryModelQuerySet(self.model, using=self._db)
        return qs.annotate(
            ce_txs_count=Count('closingentrytransactionmodel')
        )

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: EntityModel | str | UUID = None, **kwargs) -> ClosingEntryModelQuerySet:

        qs = self.get_queryset()

        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity_model=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity_model__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_model_id=entity_model)
        else:
            raise ClosingEntryValidationError(
                message='Must pass EntityModel, slug or UUID'
            )
        return qs


class ClosingEntryModelAbstract(CreateUpdateMixIn, MarkdownNotesMixIn):
    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    entity_model = models.ForeignKey('django_ledger.EntityModel',
                                     on_delete=models.CASCADE,
                                     verbose_name=_('Entity Model'))
    ledger_model = models.OneToOneField('django_ledger.LedgerModel', on_delete=models.CASCADE)
    closing_date = models.DateField(verbose_name=_('Closing Date'))
    posted = models.BooleanField(default=False, verbose_name=_('Is Posted'))
    objects = ClosingEntryModelManager.from_queryset(queryset_class=ClosingEntryModelQuerySet)()

    class Meta:
        abstract = True
        constraints = [
            models.UniqueConstraint(
                fields=[
                    'entity_model',
                    'closing_date'
                ],
                name='unique_entity_closing_date',
                violation_error_message=_('Only one Closing Entry for Date Allowed.')
            )
        ]
        indexes = [
            models.Index(fields=['entity_model']),
            models.Index(fields=['closing_date']),
            models.Index(fields=['posted']),
            models.Index(fields=['entity_model', 'posted', 'closing_date'])
        ]
        ordering = ['-closing_date']

    def __str__(self):
        return f'{self.__class__.__name__}: {self.entity_model.name} {self.closing_date}'

    def is_posted(self) -> bool:
        return self.posted is True

    def get_closing_date_as_timestamp(self):
        return make_aware(datetime.combine(self.closing_date, time.max))

    def migrate(self):

        ce_txs = self.closingentrytransactionmodel_set.all().select_related(
            'account_model',
            'unit_model'
        ).order_by(
            'tx_type',
            'account_model',
            'unit_model',
        )

        # evaluate the QS...
        ce_txs = list(ce_txs)

        # grouping by DEBIT/CREDIT
        ce_txs_gb = groupby(ce_txs, key=lambda k: k.tx_type)

        # adding DEBITS and CREDITS...
        ce_txs_sum = {
            k: sum(v.balance for v in l) for k, l in ce_txs_gb
        }

        if len(ce_txs_sum) and ce_txs_sum[TransactionModel.DEBIT] != ce_txs_sum[TransactionModel.CREDIT]:
            raise ClosingEntryValidationError(
                message=f'Invalid transactions. Credits {ce_txs_sum[TransactionModel.CREDIT]} '
                        f'do not equal Debits {ce_txs_sum[TransactionModel.DEBIT]}'
            )

        key_func = lambda i: (str(i.unit_model_id) if i.unit_model_id else '', i.activity if i.activity else '')

        ce_txs.sort(key=key_func)
        ce_txs_gb = groupby(ce_txs, key=key_func)
        ce_txs_gb = {
            unit_model_id: list(je_txs) for unit_model_id, je_txs in ce_txs_gb
        }

        ce_txs_journal_entries = {
            (unit_model_id, activity): JournalEntryModel(
                ledger_id=self.ledger_model_id,
                timestamp=self.get_closing_date_as_timestamp(),
                activity=activity if activity else None,
                entity_unit_id=unit_model_id if unit_model_id else None,
                origin='closing_entry',
                description=f'Closing Entry {self.closing_date}',
                is_closing_entry=True,
                posted=True,
                locked=True
            ) for (unit_model_id, activity), je_txs in ce_txs_gb.items()
        }

        ce_je_txs = {
            (unit_model_id, activity): [
                TransactionModel(
                    journal_entry=ce_txs_journal_entries[(unit_model_id, activity)],
                    tx_type=tx.tx_type,
                    account=tx.account_model,
                    amount=tx.balance
                ) for tx in je_txs
            ] for (unit_model_id, activity), je_txs in ce_txs_gb.items()
        }

        JournalEntryModel.objects.bulk_create(objs=chain([l for _, l in ce_txs_journal_entries.items()]))
        TransactionModel.objects.bulk_create(objs=chain.from_iterable([l for _, l in ce_je_txs.items()]))

        for k, je_model in ce_txs_journal_entries.items():
            je_model.save(verify=True)

        self.ledger_model.lock(commit=True, raise_exception=True)

        return ce_txs_journal_entries, ce_je_txs

    def create_entry_ledger(self, commit: bool = False):
        if self.ledger_model_id is None:
            ledger_model = LedgerModel(
                name=f'Closing Entry {self.closing_date} Ledger',
                entity_id=self.entity_model_id,
                hidden=True,
                locked=False,
                posted=True
            )
            ledger_model.clean()
            ledger_model.save()
            self.ledger_model = ledger_model

            if commit:
                self.save(update_fields=[
                    'ledger_model',
                    'updated'
                ])

    # ACTIONS POST...
    def can_post(self) -> bool:
        return not self.is_posted()

    def mark_as_posted(self, commit: bool = False, update_entity_meta: bool = True, **kwargs):
        if not self.can_post():
            raise ClosingEntryValidationError(
                message=_(f'Closing Entry {self.closing_date} is already posted.')
            )

        self.migrate()
        self.posted = True
        if commit:
            self.save(
                update_fields=[
                    'posted',
                    'ledger_model',
                    'updated'
                ])
            if update_entity_meta:
                self.entity_model.save_closing_entry_dates_meta(commit=True)

    def get_mark_as_posted_html_id(self) -> str:
        return f'closing_entry_post_{self.uuid}'

    def get_mark_as_posted_message(self):
        return _(f'Are you sure you want to post Closing Entry dated {self.closing_date}?')

    def get_mark_as_posted_url(self, entity_slug: Optional[str] = None) -> str:
        if not entity_slug:
            entity_slug = self.entity_model.slug
        return reverse(viewname='django_ledger:closing-entry-action-mark-as-posted',
                       kwargs={
                           'entity_slug': entity_slug,
                           'closing_entry_pk': self.uuid
                       })

    # ACTION UNPOST...
    def can_unpost(self) -> bool:
        return self.is_posted()

    def mark_as_unposted(self, commit: bool = False, update_entity_meta: bool = True, **kwargs):
        if not self.can_unpost():
            raise ClosingEntryValidationError(
                message=_(f'Closing Entry {self.closing_date} is not posted.')
            )

        self.ledger_model.unlock(commit=False, raise_exception=True)
        self.ledger_model.save(update_fields=['posted', 'locked', 'updated'])
        self.posted = False

        TransactionModel.objects.for_entity(
            entity_model=self.entity_model_id
        ).for_ledger(ledger_model=self.ledger_model).delete()

        self.ledger_model.journal_entries.all().delete()
        if commit:
            self.save(
                update_fields=[
                    'posted',
                    'ledger_model',
                    'updated'
                ])
            if update_entity_meta:
                self.entity_model.save_closing_entry_dates_meta(commit=True)

    def get_mark_as_unposted_html_id(self) -> str:
        return f'closing_entry_unpost_{self.uuid}'

    def get_mark_as_unposted_message(self):
        return _(f'Are you sure you want to unpost Closing Entry dated {self.closing_date}?')

    def get_mark_as_unposted_url(self, entity_slug: Optional[str] = None) -> str:
        if not entity_slug:
            entity_slug = self.entity_model.slug
        return reverse(viewname='django_ledger:closing-entry-action-mark-as-unposted',
                       kwargs={
                           'entity_slug': entity_slug,
                           'closing_entry_pk': self.uuid
                       })

    # ACTION CAN UPDATE TXS...
    def can_update_txs(self) -> bool:
        return not self.is_posted()

    def update_transactions(self, force_update: bool = False, post_closing_entry: bool = True, **kwargs):
        if not self.can_update_txs():
            raise ClosingEntryValidationError(
                message=_('Cannot update transactions of a posted Closing Entry.')
            )
        entity_model = self.entity_model
        entity_model.close_entity_books(
            closing_entry_model=self,
            force_update=force_update,
            post_closing_entry=post_closing_entry
        )

    def get_update_transactions_html_id(self) -> str:
        return f'closing_entry_update_txs_{self.uuid}'

    def get_update_transactions_message(self):
        return _(f'Are you sure you want to update all Closing Entry {self.closing_date} transactions? '
                 'This action will delete existing closing entry transactions and create new ones.')

    def get_update_transactions_url(self, entity_slug: Optional[str] = None) -> str:
        if not entity_slug:
            entity_slug = self.entity_model.slug
        return reverse(viewname='django_ledger:closing-entry-action-update-txs',
                       kwargs={
                           'entity_slug': entity_slug,
                           'closing_entry_pk': self.uuid
                       })

    # DELETE...
    def can_delete(self):
        return not self.is_posted()

    def delete(self, **kwargs):
        if not self.can_delete():
            raise ClosingEntryValidationError(
                message=_('Cannot delete a posted Closing Entry')
            )

        self.ledger_model.unpost(commit=True, raise_exception=True)

        TransactionModel.objects.for_entity(
            entity_model=self.entity_model_id
        ).for_ledger(ledger_model=self.ledger_model).delete()

        return self.ledger_model.delete()

    def get_delete_html_id(self) -> str:
        return f'closing_entry_delete_txs_{self.uuid}'

    def get_delete_message(self):
        return _(f'Are you sure you want to delete Closing Entry {self.closing_date}? '
                 'This action cannot be undone.')

    def get_delete_url(self, entity_slug: Optional[str] = None) -> str:
        if not entity_slug:
            entity_slug = self.entity_model.slug
        return reverse(viewname='django_ledger:closing-entry-action-delete',
                       kwargs={
                           'entity_slug': entity_slug,
                           'closing_entry_pk': self.uuid
                       })

    # HTML Tags....
    def get_html_id(self):
        return f'closing_entry_{self.uuid}'

    # URLs Generation...
    def get_list_url(self):
        return reverse(
            viewname='django_ledger:closing-entry-list',
            kwargs={
                'entity_slug': self.entity_model.slug
            }
        )


class ClosingEntryModel(ClosingEntryModelAbstract):
    class Meta(ClosingEntryModelAbstract.Meta):
        abstract = False


def closingentrymodel_presave(instance: ClosingEntryModel, **kwargs):
    instance.create_entry_ledger(commit=False)


pre_save.connect(closingentrymodel_presave, sender=ClosingEntryModel)


class ClosingEntryTransactionModelQuerySet(QuerySet):

    def for_user(self, user_model) -> 'ClosingEntryTransactionModelQuerySet':
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(closing_entry_model__entity_model__admin=user_model) |
            Q(closing_entry_model__entity_model__managers__in=[user_model])
        )

    def for_closing_date(self, closing_date: date) -> 'ClosingEntryTransactionModelQuerySet':
        return self.filter(
            closing_entry_model__closing_date__exact=closing_date
        )


class ClosingEntryTransactionModelManager(Manager):

    def get_queryset(self) -> ClosingEntryTransactionModelQuerySet:
        return ClosingEntryTransactionModelQuerySet(self.model, using=self._db)

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: EntityModel | str | UUID = None,
                   **kwargs) -> ClosingEntryTransactionModelQuerySet:
        EntityModel = lazy_loader.get_entity(entity_model)

        qs = self.get_queryset()

        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(closing_entry_model__entity_model=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(closing_entry_model__entity_model_id=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(closing_entry_model__slug__exact=entity_model)
        else:
            raise ClosingEntryValidationError(
                message=_('Must pass EntityModel or UUID or str')
            )
        return qs


class ClosingEntryTransactionModelAbstract(CreateUpdateMixIn):
    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    closing_entry_model = models.ForeignKey('django_ledger.ClosingEntryModel',
                                            on_delete=models.CASCADE)
    account_model = models.ForeignKey('django_ledger.AccountModel',
                                      on_delete=models.RESTRICT,
                                      verbose_name=_('Account Model'))
    unit_model = models.ForeignKey('django_ledger.EntityUnitModel',
                                   null=True,
                                   blank=True,
                                   on_delete=models.RESTRICT,
                                   verbose_name=_('Entity Model'))

    activity = models.CharField(max_length=20,
                                choices=JournalEntryModel.ACTIVITIES,
                                null=True,
                                blank=True,
                                verbose_name=_('Activity'))
    tx_type = models.CharField(choices=TransactionModel.TX_TYPE,
                               max_length=10,
                               verbose_name=_('Transaction Type'))
    balance = models.DecimalField(verbose_name=_('Closing Entry Balance'),
                                  max_digits=20,
                                  decimal_places=6,
                                  validators=[MinValueValidator(limit_value=Decimal('0.00'))])

    objects = ClosingEntryTransactionModelManager.from_queryset(queryset_class=ClosingEntryTransactionModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-created']
        verbose_name = _('Closing Entry Model')
        constraints = [
            models.UniqueConstraint(
                fields=[
                    'closing_entry_model',
                    'account_model',
                    'unit_model',
                    'activity'
                ],
                name='unique_closing_entry'
            ),
            models.UniqueConstraint(
                fields=[
                    'closing_entry_model',
                    'account_model',
                    'unit_model',
                ],
                condition=Q(activity=None),
                name='unique_ce_opt_1'
            ),
            models.UniqueConstraint(
                fields=[
                    'closing_entry_model',
                    'account_model',
                    'activity',
                ],
                condition=Q(unit_model=None),
                name='unique_ce_opt_2'
            ),
            models.UniqueConstraint(
                fields=[
                    'closing_entry_model',
                    'account_model'
                ],
                condition=Q(unit_model=None) & Q(activity=None),
                name='unique_ce_opt_3'
            )
        ]
        indexes = [
            models.Index(fields=['closing_entry_model']),
            models.Index(fields=['account_model'])
        ]

    def __str__(self):
        return f'{self.__class__.__name__}: {self.closing_entry_model.closing_date.strftime("%D")} | {self.balance}'

    def is_debit(self) -> bool:
        return self.tx_type == TransactionModel.DEBIT

    def is_credit(self) -> bool:
        return self.tx_type == TransactionModel.CREDIT

    def adjust_tx_type_for_negative_balance(self):
        if self.balance < Decimal('0.00'):
            if self.is_credit():
                self.tx_type = TransactionModel.DEBIT
            elif self.is_debit():
                self.tx_type = TransactionModel.CREDIT
            self.balance = abs(self.balance)

    def get_html_id(self) -> str:
        return f'closing-entry-txs-{self.uuid}'

    def clean(self):
        self.adjust_tx_type_for_negative_balance()


class ClosingEntryTransactionModel(ClosingEntryTransactionModelAbstract):
    """
    Base ClosingEntryModel Class
    """

    class Meta(ClosingEntryTransactionModelAbstract.Meta):
        abstract = False


def closingentrytransactionmodel_presave(instance: ClosingEntryTransactionModel, **kwargs):
    instance.adjust_tx_type_for_negative_balance()


pre_save.connect(closingentrytransactionmodel_presave, sender=ClosingEntryTransactionModel)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from datetime import date
from importlib import import_module
from itertools import groupby
from random import choice
from string import ascii_uppercase, ascii_lowercase, digits

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.db.models import QuerySet
from django.utils.dateparse import parse_date

from django_ledger.io.io_core import get_localdate

UserModel = get_user_model()

FAKER_IMPORTED = False

SKU_UPC_CHARS = ascii_uppercase + digits
ITEM_ID_CHARS = ascii_uppercase + ascii_lowercase + digits


def generate_random_sku(length=12):
    return ''.join(choice(SKU_UPC_CHARS) for _ in range(length))


def generate_random_upc(length=10):
    return ''.join(choice(SKU_UPC_CHARS) for _ in range(length))


def generate_random_item_id(length=20):
    return ''.join(choice(ITEM_ID_CHARS) for _ in range(length))


def get_end_date_session_key(entity_slug: str):
    return f'djl_end_date_filter_{entity_slug}'.replace('-', '_')


def get_default_entity_session_key():
    return 'djl_default_entity_model'


def get_default_unit_session_key():
    return 'djl_default_unit_model'


def set_default_entity(request, entity_model):
    session_key = get_default_entity_session_key()
    if not request.session.get(session_key):
        request.session[session_key] = {
            'entity_uuid': str(entity_model.uuid),
            'entity_slug': entity_model.slug,
            'entity_name': entity_model.name,
        }
    elif request.session[session_key].get('entity_slug') != entity_model.slug:
        request.session[session_key] = {
            'entity_uuid': str(entity_model.uuid),
            'entity_slug': entity_model.slug,
            'entity_name': entity_model.name,
        }


def get_default_entity_from_session(request):
    session_key = get_default_entity_session_key()
    return request.session.get(session_key)


def set_session_date_filter(request, entity_slug: str, end_date: date):
    session_key = get_end_date_session_key(entity_slug)
    request.session[session_key] = end_date.isoformat()


def accruable_net_summary(queryset: QuerySet) -> dict:
    """
    A convenience function that computes current net summary of accruable models.
    "net_30" group indicates the total amount is due in 30 days or less.
    "net_0" group indicates total past due amount.

    :param queryset: Accruable Objects Queryset.
    :return: A dictionary summarizing current net summary 0,30,60,90,90+ bill open amounts.
    """
    nets = {
        'net_0': 0,
        'net_30': 0,
        'net_60': 0,
        'net_90': 0,
        'net_90+': 0
    }
    nets_collect = [{
        'net_due_group': b.net_due_group(),
        'amount_open': b.get_amount_open()
    } for b in queryset]
    nets_collect.sort(key=lambda b: b['net_due_group'])
    nets_collect = {
        g: float(sum(b['amount_open'] for b in l)) for g, l in groupby(nets_collect, key=lambda b: b['net_due_group'])
    }
    nets.update(nets_collect)
    return nets


def get_end_date_from_session(entity_slug: str, request) -> date:
    session_end_date_filter = get_end_date_session_key(entity_slug)
    end_date = request.session.get(session_end_date_filter)
    end_date = parse_date(end_date) if end_date else get_localdate()
    return end_date



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>

This is a random data generator module used during the testing of the API and for Educational purposes.

The class EntityDataGenerator will only work on new entities that contain no Transactions. This is with the intention
of avoiding unintentional commingling with an actual EntityModel with production data and the data generated randomly.

This class will conveniently create a Chart of Accounts and populate the database will Bills, Invoices and various
other Transactions. The user will be able to immediately browse the Entity data by clicking on the newly created entity's
details page.

All data generated is random and fake, not related to any other entity data.
"""
from datetime import date, timedelta, datetime
from decimal import Decimal
from itertools import groupby
from random import randint, random, choice, choices
from string import ascii_uppercase
from typing import Union, Optional

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import get_localtime, get_localdate
from django_ledger.io.roles import (INCOME_OPERATIONAL, ASSET_CA_INVENTORY, COGS, ASSET_CA_CASH, ASSET_CA_PREPAID,
                                    LIABILITY_CL_DEFERRED_REVENUE, EXPENSE_OPERATIONAL, EQUITY_CAPITAL,
                                    ASSET_CA_RECEIVABLES, LIABILITY_CL_ACC_PAYABLE)
from django_ledger.models import (EntityModel, TransactionModel, VendorModel, CustomerModel,
                                  EntityUnitModel, BankAccountModel, UnitOfMeasureModel, ItemModel,
                                  BillModel, ItemTransactionModel, InvoiceModel,
                                  EstimateModel, LoggingMixIn, InvoiceModelValidationError, ChartOfAccountModel)
from django_ledger.utils import (generate_random_sku, generate_random_upc, generate_random_item_id)

try:
    from faker import Faker
    from faker.providers import company, address, phone_number, bank

    FAKER_IMPORTED = True
except ImportError:
    FAKER_IMPORTED = False


class EntityModelValidationError(ValidationError):
    pass


class EntityDataGenerator(LoggingMixIn):
    """
    A random data generator for Entity Models. Requires a user to me the entity model administrator.

    Attributes
    ----------
    user_model : UserModel
        The Django user model that administers the entity.
    entity_model : EntityModel
        The Entity model to populate.
    start_dttm: datetime
        The start datetime for new transactions. All transactions will be posted no earlier than this date.
    capital_contribution: Decimal
        The initial capital contribution amount for the Entity Model. This will help fund the entity.
    days_forward: int
        The number of days to span from the start_dttm for new transactions.

    """

    def __init__(self,
                 user_model,
                 entity_model: Union[EntityModel, str],
                 start_dttm: datetime,
                 capital_contribution: Decimal,
                 days_forward: int,
                 tx_quantity: int = 25):

        assert isinstance(entity_model, (EntityModel, str)), 'Must pass an instance of EntityModel or str'
        assert capital_contribution > 0, 'Capital contribution must be greater than 0'

        if not FAKER_IMPORTED:
            raise ImproperlyConfigured('Must install Faker library to generate random data.')

        if entity_model.admin != user_model:
            raise EntityModelValidationError(
                message=_(f'User {user_model} must have admin privileges for entity model {entity_model}.')
            )

        self.fk = Faker(['en_US'])
        self.fk.add_provider(company)
        self.fk.add_provider(address)
        self.fk.add_provider(phone_number)
        self.fk.add_provider(bank)

        self.start_date: datetime = start_dttm
        self.local_date = get_localdate()
        self.tx_quantity = tx_quantity
        self.localtime = get_localtime()
        self.COUNT_INVENTORY = True
        self.DAYS_FORWARD = days_forward

        self.entity_model: EntityModel = entity_model
        self.default_coa: Optional[ChartOfAccountModel] = None
        self.capital_contribution = capital_contribution
        self.user_model = user_model

        self.is_accruable_probability = 0.2
        self.is_paid_probability = 0.90

        self.vendor_models = None
        self.customer_models = None
        self.bank_account_models = None
        self.entity_unit_models = None
        self.uom_models = None
        self.expense_models = None
        self.product_models = None
        self.service_models = None
        self.inventory_models = None

        self.account_models = None
        self.accounts_by_role = None

        self.COUNTRY = 'US'
        self.NB_UNITS: int = 4

        self.PRODUCTS_MIN = 20
        self.PRODUCTS_MAX = 40
        self.MIN_DAYS_FORWARD = 1
        self.MAX_DAYS_FORWARD = 8

        self.logger = self.get_logger()

    def get_logger_name(self):
        return self.entity_model.slug

    def populate_entity(self, create_closing_entry: bool = False, force_populate: bool = False):

        self.logger.info('Checking for existing transactions...')
        txs_qs = TransactionModel.objects.for_entity(
            entity_model=self.entity_model
        )

        if txs_qs.count() > 0 and not force_populate:
            raise EntityModelValidationError(
                f'Cannot populate random data on {self.entity_model.name} because it already has existing Transactions'
            )

        self.create_coa()
        self.logger.info(f'Pulling Entity {self.entity_model} accounts...')
        self.account_models = self.entity_model.get_coa_accounts(order_by=('role', 'code'))
        self.accounts_by_role = {g: list(v) for g, v in groupby(self.account_models, key=lambda a: a.role)}
        self.create_vendors()
        self.create_customers()
        self.create_entity_units()
        self.create_bank_accounts()
        self.create_uom_models()

        self.create_products()
        self.create_services()
        self.create_inventories()
        self.create_expenses()

        self.fund_entity()

        for i in range(self.tx_quantity):
            start_dttm = self.start_date + timedelta(days=randint(0, self.DAYS_FORWARD))
            self.create_estimate(date_draft=start_dttm)
            self.create_po(date_draft=start_dttm)
            self.recount_inventory()
            self.update_products()
            self.create_bill(date_draft=start_dttm)

        for i in range(self.tx_quantity):
            start_dttm = self.start_date + timedelta(days=randint(0, self.DAYS_FORWARD))
            self.create_invoice(date_draft=start_dttm)

        if create_closing_entry:
            self.create_closing_entry()

    def get_next_timestamp(self, prev_timestamp: Union[date, datetime] = None) -> date:
        if not prev_timestamp:
            prev_timestamp = self.start_date

        next_timestamp = prev_timestamp + timedelta(
            days=randint(
                self.MIN_DAYS_FORWARD,
                self.MAX_DAYS_FORWARD
            ))

        if next_timestamp > self.localtime:
            next_timestamp = self.localtime
        return next_timestamp

    def create_coa(self):
        entity_model = self.entity_model

        if not self.entity_model.has_default_coa():
            coa_model = entity_model.create_chart_of_accounts(assign_as_default=True, commit=True)
        else:
            coa_model = entity_model.get_default_coa()

        entity_model.populate_default_coa(coa_model=coa_model, activate_accounts=True)
        self.default_coa = coa_model

    def create_entity_units(self, nb_units: int = None):
        self.logger.info(f'Creating entity units...')
        nb_units = self.NB_UNITS if not nb_units else nb_units

        if nb_units:
            assert nb_units >= 0, 'Number of unite must be greater than 0'

        entity_unit_models = [
            EntityUnitModel(
                name=f'Unit {u}',
                entity=self.entity_model,
                document_prefix=''.join(choices(ascii_uppercase, k=3))
            ) for u in range(nb_units)
        ]

        for unit in entity_unit_models:
            unit.clean()
            EntityUnitModel.add_root(instance=unit)

        self.entity_unit_models = self.entity_model.entityunitmodel_set.all()

    def create_vendors(self):
        self.logger.info('Creating vendors...')
        vendor_count = randint(10, 20)
        vendor_models = [
            self.entity_model.create_vendor(
                vendor_model_kwargs={
                    'vendor_name': self.fk.name() if random() > .7 else self.fk.company(),
                    'address_1': self.fk.street_address(),
                    'address_2': self.fk.building_number() if random() < .2 else None,
                    'city': self.fk.city(),
                    'state': self.fk.state_abbr(),
                    'zip_code': self.fk.postcode(),
                    'phone': self.fk.phone_number(),
                    'country': self.COUNTRY,
                    'email': self.fk.email(),
                    'website': self.fk.url(),
                    'active': True,
                    'hidden': False,
                    'description': 'A cool vendor description.'
                }, commit=False) for _ in range(vendor_count)
        ]

        for vendor in vendor_models:
            vendor.full_clean()

        self.vendor_models = VendorModel.objects.bulk_create(vendor_models, ignore_conflicts=True)

    def create_customers(self):
        self.logger.info(f'Creating entity customers...')
        customer_count = randint(10, 20)
        customer_models = [
            self.entity_model.create_customer(
                customer_model_kwargs={
                    'customer_name': self.fk.name() if random() > .2 else self.fk.company(),
                    'address_1': self.fk.street_address() + self.fk.street_suffix(),
                    'address_2': self.fk.building_number() if random() > .2 else None,
                    'city': self.fk.city(),
                    'state': self.fk.state_abbr(),
                    'zip_code': self.fk.postcode(),
                    'country': self.COUNTRY,
                    'phone': self.fk.phone_number(),
                    'email': self.fk.email(),
                    'website': self.fk.url(),
                    'active': True,
                    'hidden': False,
                    'description': f'A cool customer description. We love customers!'
                }) for _ in range(customer_count)
        ]

        for customer in customer_models:
            customer.full_clean()

        self.customer_models = CustomerModel.objects.bulk_create(customer_models, ignore_conflicts=True)

    def create_bank_accounts(self):
        self.logger.info(f'Creating entity accounts...')
        bank_account_models = [

            # creates a bank cash checking account...
            self.entity_model.create_bank_account(
                name=f'{self.entity_model.name} Checking Account',
                account_type=BankAccountModel.ACCOUNT_CHECKING,
                active=True,
                account_model=choice(self.accounts_by_role[ASSET_CA_CASH]),
                bank_account_model_kwargs={
                    'aba_number': self.fk.swift(),
                    'routing_number': str(randint(0, 9999999)).zfill(9),
                    'account_number': str(randint(0, 9999999)).zfill(9)
                },
                commit=False
            ),

            # creates a bank cash savings account...
            self.entity_model.create_bank_account(
                name=f'{self.entity_model.name} Savings Account',
                account_type=BankAccountModel.ACCOUNT_SAVINGS,
                active=True,
                account_model=choice(self.accounts_by_role[ASSET_CA_CASH]),
                bank_account_model_kwargs={
                    'aba_number': self.fk.swift(),
                    'routing_number': str(randint(0, 9999999)).zfill(9),
                    'account_number': str(randint(0, 9999999)).zfill(9)
                },
                commit=False
            )
        ]
        for ba in bank_account_models:
            ba.full_clean()

        self.bank_account_models = BankAccountModel.objects.bulk_create(bank_account_models, ignore_conflicts=True)

    def create_uom_models(self):
        self.logger.info(f'Creating entity Unit of Measures...')

        UOMs = {
            'unit': 'Unit',
            'ln-ft': 'Linear Feet',
            'sq-ft': 'Square Fee t',
            'lb': 'Pound',
            'pallet': 'Pallet',
            'man-hour': 'Man Hour'
        }

        uom_models = [
            self.entity_model.create_uom(
                unit_abbr=abbr,
                name=name,
                commit=False
            ) for abbr, name in UOMs.items()
        ]

        for uom in uom_models:
            uom.full_clean()

        self.uom_models = UnitOfMeasureModel.objects.bulk_create(uom_models)

    def create_products(self):
        self.logger.info(f'Creating entity product items...')
        product_count = randint(self.PRODUCTS_MIN, self.PRODUCTS_MAX)
        product_models = list()
        for i in range(product_count):
            # is Product....
            product_models.append(ItemModel(
                name=f'Product #{randint(1000, 9999)}',
                uom=choice(self.uom_models),
                item_role=ItemModel.ITEM_ROLE_PRODUCT,
                sku=generate_random_sku(),
                upc=generate_random_upc(),
                item_id=generate_random_item_id(),
                entity=self.entity_model,
                for_inventory=True,
                is_product_or_service=True,
                inventory_account=choice(self.accounts_by_role[ASSET_CA_INVENTORY]),
                earnings_account=choice(self.accounts_by_role[INCOME_OPERATIONAL]),
                cogs_account=choice(self.accounts_by_role[COGS]),
                additional_info=dict()
            ))

        for product in product_models:
            product.full_clean()

        ItemModel.objects.bulk_create(product_models)
        self.update_products()

    def create_services(self):
        self.logger.info(f'Creating entity service items...')
        product_count = randint(self.PRODUCTS_MIN, self.PRODUCTS_MAX)
        service_item_models = list()
        for i in range(product_count):
            service_item_models.append(ItemModel(
                name=f'Service #{randint(1000, 9999)}',
                uom=choice(self.uom_models),
                item_role=ItemModel.ITEM_ROLE_SERVICE,
                sku=generate_random_sku(),
                upc=generate_random_upc(),
                item_id=generate_random_item_id(),
                entity=self.entity_model,
                for_inventory=False,
                is_product_or_service=True,
                earnings_account=choice(self.accounts_by_role[INCOME_OPERATIONAL]),
                cogs_account=choice(self.accounts_by_role[COGS]),
                additional_info=dict()
            ))

        for service in service_item_models:
            service.full_clean()

        ItemModel.objects.bulk_create(service_item_models)
        self.update_services()

    def create_expenses(self):
        self.logger.info(f'Creating entity expense items...')
        expense_count = randint(self.PRODUCTS_MIN, self.PRODUCTS_MAX)
        expense_models = [
            ItemModel(
                name=f'Expense Item {randint(1000, 9999)}',
                uom=choice(self.uom_models),
                item_type=choice(ItemModel.ITEM_TYPE_CHOICES)[0],
                item_role=ItemModel.ITEM_ROLE_EXPENSE,
                sku=generate_random_sku(),
                upc=generate_random_upc(),
                item_id=generate_random_item_id(),
                entity=self.entity_model,
                is_product_or_service=False,
                for_inventory=False,
                expense_account=choice(self.accounts_by_role[EXPENSE_OPERATIONAL]),
            ) for _ in range(expense_count)
        ]

        for exp in expense_models:
            exp.full_clean()

        ItemModel.objects.bulk_create(expense_models)
        self.update_expenses()

    def create_inventories(self):
        self.logger.info(f'Creating entity inventory items...')
        inv_count = randint(self.PRODUCTS_MIN, self.PRODUCTS_MAX)
        inventory_models = [
            ItemModel(
                name=f'Inventory {randint(1000, 9999)}',
                uom=choice(self.uom_models),
                item_role=ItemModel.ITEM_ROLE_INVENTORY,
                item_type=choice(ItemModel.ITEM_TYPE_CHOICES)[0],
                item_id=generate_random_item_id(),
                entity=self.entity_model,
                for_inventory=True,
                is_product_or_service=True if random() > 0.6 else False,
                sku=generate_random_sku(),
                upc=generate_random_upc(),
                earnings_account=choice(self.accounts_by_role[INCOME_OPERATIONAL]),
                cogs_account=choice(self.accounts_by_role[COGS]),
                inventory_account=choice(self.accounts_by_role[ASSET_CA_INVENTORY]),
            ) for _ in range(inv_count)
        ]

        for inv in inventory_models:
            inv.full_clean()

        self.inventory_models = ItemModel.objects.bulk_create(inventory_models)

    def update_products(self):
        self.logger.info(f'Updating product catalog...')
        self.product_models = self.entity_model.get_items_products()

    def update_services(self):
        self.logger.info(f'Updating service catalog...')
        self.service_models = self.entity_model.get_items_services()

    def update_inventory(self):
        self.logger.info(f'Updating inventory...')
        self.inventory_models = self.entity_model.get_items_inventory()

    def update_expenses(self):
        self.logger.info(f'Updating expenses...')
        self.expense_models = self.entity_model.get_items_expenses()

    def create_estimate(self, date_draft: date):
        estimate_model = self.entity_model.create_estimate(
            estimate_title=f'Customer Estimate {date_draft}',
            date_draft=date_draft,
            customer_model=choice(self.customer_models),
            contract_terms=choice(EstimateModel.CONTRACT_TERMS_CHOICES_VALID),
            commit=True
        )
        self.logger.info(f'Creating entity estimate {estimate_model.estimate_number}...')

        estimate_items = [
            ItemTransactionModel(
                ce_model=estimate_model,
                item_model=choice(self.product_models),
                ce_quantity=round(random() * randint(5, 15), 2),
                ce_unit_cost_estimate=round(random() * randint(50, 100), 2),
                ce_unit_revenue_estimate=round(random() * randint(80, 120) * (1 + 0.2 * random()), 2),
                entity_unit=choice(self.entity_unit_models) if random() > .75 else None
            ) for _ in range(randint(1, 10))
        ]

        for i in estimate_items:
            i.full_clean()

        estimate_model.full_clean()
        estimate_model.update_state(itemtxs_qs=estimate_items)
        estimate_model.save()

        estimate_items = estimate_model.itemtransactionmodel_set.bulk_create(objs=estimate_items)

        if random() > 0.25:
            date_in_review = self.get_next_timestamp(date_draft)
            estimate_model.mark_as_review(commit=True, date_in_review=date_in_review)
            if random() > 0.50:
                date_approved = self.get_next_timestamp(date_in_review)
                estimate_model.mark_as_approved(commit=True, date_approved=date_approved)
                if random() > 0.25:
                    date_completed = self.get_next_timestamp(date_approved)
                    estimate_model.mark_as_completed(commit=True, date_completed=date_completed)
                elif random() > 0.8:
                    date_void = self.get_next_timestamp(date_approved)
                    estimate_model.mark_as_void(commit=True, date_void=date_void)
            elif random() > 0.8:
                date_canceled = self.get_next_timestamp(date_in_review)
                estimate_model.mark_as_canceled(commit=True, date_canceled=date_canceled)

    def create_bill(self, date_draft: date):
        bill_model = self.entity_model.create_bill(
            vendor_model=choice(self.vendor_models),
            cash_account=choice(self.accounts_by_role[ASSET_CA_CASH]),
            prepaid_account=choice(self.accounts_by_role[ASSET_CA_PREPAID]),
            payable_account=choice(self.accounts_by_role[LIABILITY_CL_ACC_PAYABLE]),
            terms=choice(BillModel.TERM_CHOICES_VALID),
            date_draft=date_draft,
            additional_info=dict(),
            commit=True
        )

        self.logger.info(f'Creating entity bill {bill_model.bill_number}...')

        bill_items = [
            ItemTransactionModel(
                bill_model=bill_model,
                item_model=choice(self.expense_models),
                quantity=round(random() * randint(5, 15), 2),
                unit_cost=round(random() * randint(50, 100), 2),
                entity_unit=choice(self.entity_unit_models) if random() > .75 else None
            ) for _ in range(randint(1, 10))
        ]

        for bi in bill_items:
            bi.full_clean()

        bill_model.update_amount_due(itemtxs_qs=bill_items)
        bill_model.itemtransactionmodel_set.bulk_create(bill_items)
        bill_model.full_clean()
        bill_model.save()

        if random() > 0.25 and bill_model.amount_due:
            date_in_review = self.get_next_timestamp(date_draft)
            bill_model.mark_as_review(commit=True, date_in_review=date_in_review)

            if random() > 0.50:
                date_approved = self.get_next_timestamp(date_in_review)
                bill_model.mark_as_approved(
                    commit=True,
                    entity_slug=self.entity_model.slug,
                    user_model=self.user_model,
                    date_approved=date_approved
                )

                if random() > 0.25:
                    paid_date = self.get_next_timestamp(date_approved)
                    bill_model.mark_as_paid(
                        user_model=self.user_model,
                        entity_slug=self.entity_model.slug,
                        date_paid=paid_date,
                        commit=True
                    )
                elif random() > 0.8:
                    void_date = self.get_next_timestamp(date_approved)
                    bill_model.mark_as_void(
                        user_model=self.user_model,
                        entity_slug=self.entity_model.slug,
                        date_void=void_date,
                        commit=True
                    )
            elif random() > 0.8:
                canceled_date = self.get_next_timestamp(date_in_review)
                bill_model.mark_as_canceled(date_canceled=canceled_date)

    def create_po(self, date_draft: date):

        po_model = self.entity_model.create_purchase_order(date_draft=date_draft)

        po_items = [
            ItemTransactionModel(
                po_model=po_model,
                item_model=choice(self.product_models),
                po_quantity=round(random() * randint(3, 10) + 3, 2),
                po_unit_cost=round(random() * randint(100, 800), 2),
                entity_unit=choice(self.entity_unit_models) if random() > .75 else None
            ) for _ in range(randint(1, 10))
        ]

        for poi in po_items:
            poi.full_clean()

        self.logger.info(f'Creating entity purchase order {po_model.po_number}...')
        po_items = po_model.itemtransactionmodel_set.bulk_create(po_items)
        po_model.update_state(itemtxs_qs=po_items)
        po_model.full_clean()
        po_model.save()

        # mark as approved...
        if random() > 0.25 and po_model.po_amount:
            date_review = self.get_next_timestamp(date_draft)
            po_model.mark_as_review(commit=True, date_in_review=date_review)
            if random() > 0.5:
                date_approved = self.get_next_timestamp(date_review)
                po_model.mark_as_approved(commit=True, date_approved=date_approved)
                if random() > 0.25:
                    # add a PO bill...
                    date_fulfilled = self.get_next_timestamp(date_approved)
                    date_bill_draft = date_fulfilled - timedelta(days=randint(1, 3))

                    bill_model = self.entity_model.create_bill(
                        vendor_model=choice(self.vendor_models),
                        terms=choice(BillModel.TERM_CHOICES_VALID),
                        date_draft=date_bill_draft,
                        cash_account=choice(self.accounts_by_role[ASSET_CA_CASH]),
                        prepaid_account=choice(self.accounts_by_role[ASSET_CA_PREPAID]),
                        payable_account=choice(self.accounts_by_role[LIABILITY_CL_ACC_PAYABLE]),
                        commit=True
                    )

                    for po_i in po_items:
                        po_i.po_total_amount = round(po_i.po_total_amount, 2)
                        po_i.total_amount = round(po_i.po_total_amount, 2)
                        po_i.quantity = round(po_i.po_quantity, 2)
                        po_i.unit_cost = round(po_i.po_unit_cost, 2)
                        po_i.bill_model = bill_model
                        po_i.po_item_status = ItemTransactionModel.STATUS_RECEIVED
                        po_i.full_clean()

                    bill_model.update_amount_due(itemtxs_qs=po_items)
                    bill_model.full_clean()
                    bill_model.update_state()
                    bill_model.save()

                    po_model.itemtransactionmodel_set.bulk_update(
                        po_items,
                        fields=[
                            'po_total_amount',
                            'total_amount',
                            'po_quantity',
                            'quantity',
                            'po_unit_cost',
                            'unit_cost',
                            'bill_model',
                            'po_item_status'
                        ])

                    if random() > 0.25:
                        date_bill_review = self.get_next_timestamp(date_bill_draft)
                        bill_model.mark_as_review(commit=True, date_in_review=date_bill_review)
                        if random() > 0.50:
                            bill_approve_date = self.get_next_timestamp(date_bill_review)
                            bill_model.mark_as_approved(commit=True,
                                                        entity_slug=self.entity_model.slug,
                                                        user_model=self.user_model,
                                                        date_approved=bill_approve_date)
                            if random() > 0.25:
                                bill_paid_date = self.get_next_timestamp(bill_approve_date)
                                bill_model.mark_as_paid(
                                    user_model=self.user_model,
                                    entity_slug=self.entity_model.slug,
                                    commit=True,
                                    date_paid=bill_paid_date)

                                if random() > 0.20:
                                    for po_i in po_items:
                                        po_i.po_item_status = ItemTransactionModel.STATUS_RECEIVED
                                        po_i.po_item_status = ItemTransactionModel.STATUS_RECEIVED
                                        po_i.full_clean()

                                    # todo: can pass po items??..
                                    po_model.itemtransactionmodel_set.bulk_update(po_items,
                                                                                  fields=[
                                                                                      'po_item_status',
                                                                                      'updated'
                                                                                  ])
                                    po_model.mark_as_fulfilled(
                                        date_fulfilled=date_fulfilled,
                                        commit=True)

                                    self.entity_model.update_inventory(
                                        # user_model=self.user_model,
                                        commit=True)

                                    self.update_products()
                                    self.update_inventory()

    def create_invoice(self, date_draft: date):
        invoice_model = self.entity_model.create_invoice(
            customer_model=choice(self.customer_models),
            terms=choice(InvoiceModel.TERM_CHOICES_VALID),
            cash_account=choice(self.accounts_by_role[ASSET_CA_CASH]),
            prepaid_account=choice(self.accounts_by_role[ASSET_CA_RECEIVABLES]),
            payable_account=choice(self.accounts_by_role[LIABILITY_CL_DEFERRED_REVENUE]),
            date_draft=date_draft,
            additional_info=dict(),
            commit=True
        )
        self.logger.info(f'Creating entity invoice {invoice_model.invoice_number}...')

        invoice_items = list()

        for i in range(randint(1, 10)):
            item_model: ItemModel = choice(self.product_models)
            quantity = Decimal.from_float(round(random() * randint(1, 2), 2))
            entity_unit = choice(self.entity_unit_models) if random() > .75 else None
            margin = Decimal(random() + 3.5)
            avg_cost = item_model.get_average_cost()
            if item_model.is_product():
                if item_model.inventory_received is not None and item_model.inventory_received > 0.0:
                    if quantity > item_model.inventory_received:
                        quantity = item_model.inventory_received

                    # reducing inventory qty...
                    item_model.inventory_received -= quantity
                    item_model.inventory_received_value -= avg_cost * quantity
                    unit_cost = avg_cost * margin
                else:
                    quantity = 0.0
                    unit_cost = 0.0

                if all([
                    quantity > 0.00,
                    unit_cost > 0.00
                ]):
                    itm = ItemTransactionModel(
                        invoice_model=invoice_model,
                        item_model=item_model,
                        quantity=quantity,
                        unit_cost=unit_cost,
                        entity_unit=entity_unit
                    )
                    itm.full_clean()
                    invoice_items.append(itm)

        invoice_items = invoice_model.itemtransactionmodel_set.bulk_create(invoice_items)
        invoice_model.update_amount_due(itemtxs_qs=invoice_items)
        invoice_model.full_clean()
        invoice_model.save()

        if random() > 0.25 and invoice_model.amount_due:
            date_review = self.get_next_timestamp(date_draft)

            try:
                invoice_model.mark_as_review(commit=True, date_in_review=date_review)
            except InvoiceModelValidationError as e:
                # invoice cannot be marked as in review...
                return

            if random() > 0.50:
                date_approved = self.get_next_timestamp(date_review)
                invoice_model.mark_as_approved(entity_slug=self.entity_model.slug,
                                               user_model=self.user_model,
                                               commit=True,
                                               date_approved=date_approved)
                if random() > 0.25:
                    date_paid = self.get_next_timestamp(date_approved)
                    invoice_model.mark_as_paid(
                        entity_slug=self.entity_model.slug,
                        user_model=self.user_model,
                        date_paid=date_paid,
                        commit=True
                    )
                    self.entity_model.update_inventory(
                        # user_model=self.user_model,
                        commit=True
                    )
                    self.update_inventory()
                    self.update_products()
                elif random() > 0.8:
                    date_void = self.get_next_timestamp(date_approved)
                    invoice_model.mark_as_void(
                        entity_slug=self.entity_model.slug,
                        user_model=self.user_model,
                        date_void=date_void,
                        commit=True
                    )
            elif random() > 0.8:
                date_canceled = self.get_next_timestamp(date_review)
                invoice_model.mark_as_canceled(commit=True, date_canceled=date_canceled)

    def fund_entity(self):

        self.logger.info(f'Funding entity...')
        capital_acc = choice(self.accounts_by_role[EQUITY_CAPITAL])
        cash_acc = choice(self.bank_account_models).account_model

        self.entity_model.deposit_capital(
            cash_account=cash_acc,
            capital_account=capital_acc,
            amount=self.capital_contribution,
            je_timestamp=self.start_date,
            je_posted=True,
            ledger_posted=True,
            description='Entity Funding for Sample Data',
        )

    def create_closing_entry(self):
        closing_date = self.start_date + timedelta(days=int(self.DAYS_FORWARD / 2))
        ce_model, ce_txs = self.entity_model.close_books_for_month(
            year=closing_date.year,
            month=closing_date.month,
            post_closing_entry=True
        )

    def recount_inventory(self):
        self.logger.info(f'Recounting inventory...')
        self.entity_model.update_inventory(
            # user_model=self.user_model,
            commit=True
        )



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from django.contrib import messages
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.http import HttpResponseRedirect, HttpResponseNotFound, HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import (CreateView, ArchiveIndexView, YearArchiveView, MonthArchiveView, DetailView,
                                  UpdateView, DeleteView, RedirectView)
from django.views.generic.detail import SingleObjectMixin

from django_ledger.forms.purchase_order import (PurchaseOrderModelCreateForm, BasePurchaseOrderModelUpdateForm,
                                                DraftPurchaseOrderModelUpdateForm, ReviewPurchaseOrderModelUpdateForm,
                                                ApprovedPurchaseOrderModelUpdateForm,
                                                get_po_itemtxs_formset_class)
from django_ledger.models import PurchaseOrderModel, ItemTransactionModel, EstimateModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class PurchaseOrderModelModelViewQuerySetMixIn(DjangoLedgerSecurityMixIn):
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = PurchaseOrderModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL,
            ).select_related('entity', 'ce_model')
        return super().get_queryset()


class PurchaseOrderModelListView(PurchaseOrderModelModelViewQuerySetMixIn, ArchiveIndexView):
    template_name = 'django_ledger/purchase_order/po_list.html'
    context_object_name = 'po_list'
    PAGE_TITLE = _('PO List')
    date_field = 'created'
    paginate_by = 10
    paginate_orphans = 2
    allow_empty = True
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'uil:bill'
    }

    def get_allow_future(self):
        allow_future = self.request.GET.get('allow_future')
        if allow_future:
            try:
                allow_future = int(allow_future)
                if allow_future in (0, 1):
                    return bool(allow_future)
            except ValueError:
                pass
        return False


class PurchaseOrderModelYearListView(YearArchiveView,
                                     PurchaseOrderModelListView):
    paginate_by = 10
    make_object_list = True


class PurchaseOrderModelMonthListView(MonthArchiveView,
                                      PurchaseOrderModelListView):
    paginate_by = 10
    month_format = '%m'
    date_list_period = 'year'


class PurchaseOrderModelCreateView(PurchaseOrderModelModelViewQuerySetMixIn, CreateView):
    template_name = 'django_ledger/purchase_order/po_create.html'
    PAGE_TITLE = _('Create Purchase Order')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'uil:bill'
    }
    for_estimate = False

    def get(self, request, entity_slug, **kwargs):
        response = super(PurchaseOrderModelCreateView, self).get(request, entity_slug, **kwargs)
        if self.for_estimate and 'ce_pk' in self.kwargs:
            estimate_qs = EstimateModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL,
            )
            estimate_model: EstimateModel = get_object_or_404(estimate_qs, uuid__exact=self.kwargs['ce_pk'])
            if not estimate_model.can_bind():
                return HttpResponseNotFound('404 Not Found')
        return response

    def get_context_data(self, **kwargs):
        context = super(PurchaseOrderModelCreateView, self).get_context_data(**kwargs)

        if self.for_estimate:
            context['form_action_url'] = reverse('django_ledger:po-create-estimate',
                                                 kwargs={
                                                     'entity_slug': self.kwargs['entity_slug'],
                                                     'ce_pk': self.kwargs['ce_pk']
                                                 })
            estimate_qs = EstimateModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL
            ).select_related('customer')
            estimate_model = get_object_or_404(estimate_qs, uuid__exact=self.kwargs['ce_pk'])
            context['estimate_model'] = estimate_model

        else:
            context['form_action_url'] = reverse('django_ledger:po-create',
                                                 kwargs={
                                                     'entity_slug': self.kwargs['entity_slug']
                                                 })
        return context

    def get_form(self, form_class=None):
        entity_slug = self.kwargs['entity_slug']
        form = PurchaseOrderModelCreateForm(
            entity_slug=entity_slug,
            user_model=self.request.user,
            **self.get_form_kwargs()
        )
        return form

    def form_valid(self, form):
        po_model: PurchaseOrderModel = form.save(commit=False)
        po_model = po_model.configure(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user
        )

        if self.for_estimate:
            ce_pk = self.kwargs['ce_pk']
            estimate_model_qs = EstimateModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL,
            ).for_user(user_model=self.request.user)
            estimate_model = get_object_or_404(estimate_model_qs, uuid__exact=ce_pk)
            po_model.action_bind_estimate(estimate_model=estimate_model, commit=False)
        return super().form_valid(form=form)

    def get_success_url(self):
        entity_slug = self.kwargs['entity_slug']

        if self.for_estimate:
            ce_pk = self.kwargs['ce_pk']
            return reverse('django_ledger:customer-estimate-detail',
                           kwargs={
                               'entity_slug': self.kwargs['entity_slug'],
                               'ce_pk': ce_pk
                           })
        return reverse('django_ledger:po-list',
                       kwargs={
                           'entity_slug': entity_slug
                       })


class PurchaseOrderModelUpdateView(PurchaseOrderModelModelViewQuerySetMixIn, UpdateView):
    slug_url_kwarg = 'po_pk'
    slug_field = 'uuid'
    context_object_name = 'po_model'
    template_name = 'django_ledger/purchase_order/po_update.html'
    extra_context = {
        'header_subtitle_icon': 'uil:bill'
    }
    action_update_items = False

    def get_context_data(self, itemtxs_formset=None, **kwargs):
        context = super().get_context_data(**kwargs)
        po_model: PurchaseOrderModel = self.object
        title = f'Purchase Order {po_model.po_number}'
        context['page_title'] = title
        context['header_title'] = title

        # continue here: pass item QS to item_form...
        if not itemtxs_formset:
            itemtxs_qs = self.get_po_itemtxs_qs(po_model)
            itemtxs_qs, itemtxs_agg = po_model.get_itemtxs_data(queryset=itemtxs_qs)
            po_itemtxs_formset_class = get_po_itemtxs_formset_class(po_model)
            itemtxs_formset = po_itemtxs_formset_class(
                entity_slug=self.kwargs['entity_slug'],
                user_model=self.request.user,
                po_model=po_model,
                queryset=itemtxs_qs,
            )
        else:
            itemtxs_qs, itemtxs_agg = po_model.get_itemtxs_data()

        context['po_total_amount__sum'] = itemtxs_agg['po_total_amount__sum']
        context['bill_amount_paid__sum'] = itemtxs_agg['bill_amount_paid__sum']
        context['itemtxs_qs'] = itemtxs_qs
        context['itemtxs_formset'] = itemtxs_formset

        if po_model.is_contract_bound():
            ce_model: EstimateModel = po_model.ce_model
            ce_itemtxs_qs, ce_itemtxs_agg = ce_model.get_itemtxs_annotation()
            context['ce_itemtxs_agg'] = ce_itemtxs_agg
            context['ce_cost_estimate__sum'] = sum(i['ce_cost_estimate__sum'] for i in ce_itemtxs_agg)

        return context

    def get(self, request, entity_slug, po_pk, *args, **kwargs):
        if self.action_update_items:
            return HttpResponseRedirect(
                redirect_to=reverse('django_ledger:po-update',
                                    kwargs={
                                        'entity_slug': entity_slug,
                                        'po_pk': po_pk
                                    })
            )
        return super(PurchaseOrderModelUpdateView, self).get(request, entity_slug, po_pk, *args, **kwargs)

    def post(self, request, entity_slug, *args, **kwargs):
        if self.action_update_items:

            if not request.user.is_authenticated:
                return HttpResponseForbidden()

            queryset = self.get_queryset()
            po_model: PurchaseOrderModel = self.get_object(queryset=queryset)
            self.object = po_model
            po_itemtxs_formset_class = get_po_itemtxs_formset_class(po_model)
            itemtxs_formset = po_itemtxs_formset_class(request.POST,
                                                       user_model=self.request.user,
                                                       po_model=po_model,
                                                       entity_slug=entity_slug)

            if itemtxs_formset.has_changed():
                if itemtxs_formset.is_valid():
                    itemtxs_list = itemtxs_formset.save(commit=False)
                    create_bill_uuids = [
                        str(i['uuid'].uuid) for i in itemtxs_formset.cleaned_data if i and i['create_bill'] is True
                    ]

                    if create_bill_uuids:
                        item_uuids = ','.join(create_bill_uuids)
                        redirect_url = reverse(
                            'django_ledger:bill-create-po',
                            kwargs={
                                'entity_slug': self.kwargs['entity_slug'],
                                'po_pk': po_model.uuid,
                            }
                        )
                        redirect_url += f'?item_uuids={item_uuids}'
                        return HttpResponseRedirect(redirect_url)

                    for itemtxs in itemtxs_list:
                        if not itemtxs.po_model_id:
                            itemtxs.po_model_id = po_model.uuid
                        itemtxs.clean()

                    itemtxs_list = itemtxs_formset.save()
                    po_model.update_state()
                    po_model.clean()
                    po_model.save(update_fields=['po_amount',
                                                 'po_amount_received',
                                                 'updated'])
                    # if valid get saved formset from DB
                    return self.render_to_response(context=self.get_context_data())
                # if not valid, return formset with errors...
                return self.render_to_response(context=self.get_context_data(itemtxs_formset=itemtxs_formset))
        return super(PurchaseOrderModelUpdateView, self).post(request, **kwargs)

    def get_form(self, form_class=None):
        po_model: PurchaseOrderModel = self.object
        if po_model.is_draft():
            return DraftPurchaseOrderModelUpdateForm(
                entity_slug=self.kwargs['entity_slug'],
                user_model=self.request.user,
                **self.get_form_kwargs()
            )
        elif po_model.is_review():
            return ReviewPurchaseOrderModelUpdateForm(
                entity_slug=self.kwargs['entity_slug'],
                user_model=self.request.user,
                **self.get_form_kwargs()
            )
        elif po_model.is_approved():
            return ApprovedPurchaseOrderModelUpdateForm(
                entity_slug=self.kwargs['entity_slug'],
                user_model=self.request.user,
                **self.get_form_kwargs()
            )
        return BasePurchaseOrderModelUpdateForm(
            entity_slug=self.kwargs['entity_slug'],
            user_model=self.request.user,
            **self.get_form_kwargs()
        )

    def get_form_kwargs(self):
        if self.action_update_items:
            return {
                'initial': self.get_initial(),
                'prefix': self.get_prefix(),
                'instance': self.object
            }
        return super(PurchaseOrderModelUpdateView, self).get_form_kwargs()

    def get_success_url(self):
        entity_slug = self.kwargs['entity_slug']
        po_pk = self.kwargs['po_pk']
        return reverse('django_ledger:po-detail',
                       kwargs={
                           'entity_slug': entity_slug,
                           'po_pk': po_pk
                       })

    def get_po_itemtxs_qs(self, po_model: PurchaseOrderModel):
        return po_model.itemtransactionmodel_set.select_related('bill_model', 'po_model').order_by('created')

    def form_valid(self, form: BasePurchaseOrderModelUpdateForm):
        po_model: PurchaseOrderModel = form.save(commit=False)

        if form.has_changed():
            po_items_qs = ItemTransactionModel.objects.for_po(
                entity_model=self.kwargs['entity_slug'],
                po_pk=po_model.uuid,
            ).select_related('bill_model')

            if all(['po_status' in form.changed_data,
                    po_model.po_status == po_model.PO_STATUS_APPROVED]):
                po_items_qs.update(po_item_status=ItemTransactionModel.STATUS_NOT_ORDERED)

            if 'fulfilled' in form.changed_data:

                if not all([i.bill_model for i in po_items_qs]):
                    messages.add_message(self.request,
                                         messages.ERROR,
                                         f'All PO items must be billed before marking'
                                         f' PO: {po_model.po_number} as fulfilled.',
                                         extra_tags='is-danger')
                    return self.get(self.request)

                else:
                    if not all([i.bill_model.is_paid() for i in po_items_qs]):
                        messages.add_message(self.request,
                                             messages.SUCCESS,
                                             f'All bills must be paid before marking'
                                             f' PO: {po_model.po_number} as fulfilled.',
                                             extra_tags='is-success')
                        return self.get(self.request)

                po_items_qs.update(po_item_status=ItemTransactionModel.STATUS_RECEIVED)

        messages.add_message(self.request,
                             messages.SUCCESS,
                             f'{self.object.po_number} successfully updated.',
                             extra_tags='is-success')

        return super().form_valid(form)


class PurchaseOrderModelDetailView(PurchaseOrderModelModelViewQuerySetMixIn, DetailView):
    slug_url_kwarg = 'po_pk'
    slug_field = 'uuid'
    context_object_name = 'po_model'
    template_name = 'django_ledger/purchase_order/po_detail.html'
    extra_context = {
        'header_subtitle_icon': 'uil:bill',
        'hide_menu': True
    }

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        po_model: PurchaseOrderModel = self.object
        title = f'Purchase Order {po_model.po_number}'
        context['page_title'] = title
        context['header_title'] = title

        po_model: PurchaseOrderModel = self.object
        po_items_qs, item_data = po_model.get_itemtxs_data(
            queryset=po_model.itemtransactionmodel_set.all().select_related('item_model', 'bill_model')
        )
        context['po_items'] = po_items_qs
        context['po_total_amount'] = sum(
            i['po_total_amount'] for i in po_items_qs.values(
                'po_total_amount', 'po_item_status') if i['po_item_status'] != 'cancelled')
        return context


class PurchaseOrderModelDeleteView(PurchaseOrderModelModelViewQuerySetMixIn, DeleteView):
    slug_url_kwarg = 'po_pk'
    slug_field = 'uuid'
    context_object_name = 'po_model'
    template_name = 'django_ledger/purchase_order/po_delete.html'
    extra_context = {
        'hide_menu': True,
        'header_subtitle_icon': 'uil:bill'
    }

    def get_context_data(self, *, object_list=None, **kwargs):
        context = super().get_context_data(object_list=object_list, **kwargs)
        po_model: PurchaseOrderModel = self.object
        context['page_title'] = _('Delete Purchase Order ') + po_model.po_number
        context['header_title'] = context['page_title']
        return context

    def get_success_url(self):
        return reverse('django_ledger:entity-dashboard',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug'],
                       })

    def form_valid(self, request, *args, **kwargs):
        po_model: PurchaseOrderModel = self.get_object()
        self.object = po_model
        po_items_qs = po_model.itemtransactionmodel_set.filter(bill_model__isnull=False)
        if po_items_qs.exists():
            messages.add_message(request,
                                 message=f'Cannot delete {po_model.po_number} because it has related bills.',
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
            url = reverse('django_ledger:po-update',
                          kwargs={
                              'entity_slug': self.kwargs['entity_slug'],
                              'po_pk': self.kwargs['po_pk']
                          })
            return HttpResponseRedirect(url)
        success_url = self.get_success_url()
        self.object.delete()
        return HttpResponseRedirect(success_url)


# ACTIONS...
class BasePurchaseOrderActionActionView(
    PurchaseOrderModelModelViewQuerySetMixIn,
    RedirectView,
    SingleObjectMixin
):
    http_method_names = ['get']
    pk_url_kwarg = 'po_pk'
    action_name = None
    commit = True

    def get_redirect_url(self, entity_slug, po_pk, *args, **kwargs):
        return reverse('django_ledger:po-update',
                       kwargs={
                           'entity_slug': entity_slug,
                           'po_pk': po_pk
                       })

    def get(self, request, *args, **kwargs):
        kwargs['user_model'] = self.request.user
        if not self.action_name:
            raise ImproperlyConfigured('View attribute action_name is required.')
        response = super(BasePurchaseOrderActionActionView, self).get(request, *args, **kwargs)
        po_model: PurchaseOrderModel = self.get_object()

        try:
            getattr(po_model, self.action_name)(commit=self.commit, **kwargs)
        except ValidationError as e:
            messages.add_message(request,
                                 message=e.message,
                                 level=messages.ERROR,
                                 extra_tags='is-danger')
        return response


class PurchaseOrderMarkAsDraftView(BasePurchaseOrderActionActionView):
    action_name = 'mark_as_draft'


class PurchaseOrderMarkAsReviewView(BasePurchaseOrderActionActionView):
    action_name = 'mark_as_review'


class PurchaseOrderMarkAsApprovedView(BasePurchaseOrderActionActionView):
    action_name = 'mark_as_approved'


class PurchaseOrderMarkAsFulfilledView(BasePurchaseOrderActionActionView):
    action_name = 'mark_as_fulfilled'


class PurchaseOrderMarkAsCanceledView(BasePurchaseOrderActionActionView):
    action_name = 'mark_as_canceled'


class PurchaseOrderMarkAsVoidView(BasePurchaseOrderActionActionView):
    action_name = 'mark_as_void'



# Generated by Django 5.2.1 on 2025-08-26 12:40

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0023_customermodel_customer_code_customermodel_picture_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='billmodel',
            name='entity_model',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel'),
        ),
        migrations.AddField(
            model_name='invoicemodel',
            name='entity_model',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_ledger.entitymodel'),
        ),
    ]



from django.urls import path

from django_ledger import views

urlpatterns = [
    path('<slug:entity_slug>/unit/list/',
         views.EntityUnitModelListView.as_view(),
         name='unit-list'),
    path('<slug:entity_slug>/detail/<slug:unit_slug>/',
         views.EntityUnitModelDetailView.as_view(),
         name='unit-detail'),
    path('<slug:entity_slug>/unit/create/',
         views.EntityUnitModelCreateView.as_view(),
         name='unit-create'),
    path('<slug:entity_slug>/unit/update/<slug:unit_slug>/',
         views.EntityUnitUpdateView.as_view(),
         name='unit-update'),


    # DASHBOARD Views ...
    path('<slug:entity_slug>/dashboard/<slug:unit_slug>/',
         views.EntityModelDetailHandlerView.as_view(),
         name='unit-dashboard'),
    path('<slug:entity_slug>/dashboard/<slug:unit_slug>/year/<int:year>/',
         views.FiscalYearEntityModelDashboardView.as_view(),
         name='unit-dashboard-year'),
    path('<slug:entity_slug>/dashboard/<slug:unit_slug>/quarter/<int:year>/<int:quarter>/',
         views.QuarterlyEntityDashboardView.as_view(),
         name='unit-dashboard-quarter'),
    path('<slug:entity_slug>/dashboard/<slug:unit_slug>/month/<int:year>/<int:month>/',
         views.MonthlyEntityDashboardView.as_view(),
         name='unit-dashboard-month'),
    path('<slug:entity_slug>/dashboard/<slug:unit_slug>/date/<int:year>/<int:month>/<int:day>/',
         views.DateEntityDashboardView.as_view(),
         name='unit-dashboard-date'),

]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from datetime import datetime, time

from django.contrib import messages
from django.urls import reverse
from django.utils.timezone import make_aware
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, FormView, DetailView, UpdateView, DeleteView

from django_ledger.forms.data_import import ImportJobModelCreateForm, ImportJobModelUpdateForm
from django_ledger.forms.data_import import StagedTransactionModelFormSet
from django_ledger.io.ofx import OFXFileManager
from django_ledger.models.data_import import ImportJobModel, StagedTransactionModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class ImportJobModelViewBaseView(DjangoLedgerSecurityMixIn):
    queryset = None

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = ImportJobModel.objects.for_entity(
                entity_model=self.AUTHORIZED_ENTITY_MODEL,
            ).order_by('-created').select_related(
                'bank_account_model',
                'bank_account_model__entity_model',
                'bank_account_model__account_model',
                'bank_account_model__account_model__coa_model'
            )
        return self.queryset


class ImportJobModelCreateView(ImportJobModelViewBaseView, FormView):
    template_name = 'django_ledger/data_import/import_job_create.html'
    PAGE_TITLE = _('Create Import Job')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }
    form_class = ImportJobModelCreateForm

    def get_form(self, form_class=None, **kwargs):
        return self.form_class(
            entity_model=self.get_authorized_entity_instance(),
            **self.get_form_kwargs()
        )

    def get_success_url(self):
        return reverse('django_ledger:data-import-jobs-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })

    def form_valid(self, form):
        ofx_manager = OFXFileManager(ofx_file_or_path=form.files['ofx_file'])
        import_job: ImportJobModel = form.save(commit=False)
        import_job.configure(commit=False)
        import_job.save()

        txs_to_stage = ofx_manager.get_account_txs()
        staged_txs_model_list = [
            StagedTransactionModel(
                date_posted=make_aware(value=datetime.combine(date=tx.dtposted.date(), time=time.min)),
                fit_id=tx.fitid,
                amount=tx.trnamt,
                import_job=import_job,
                name=tx.name,
                memo=tx.memo
            ) for tx in txs_to_stage
        ]
        for tx in staged_txs_model_list:
            tx.clean()

        StagedTransactionModel.objects.bulk_create(staged_txs_model_list)
        return super().form_valid(form=form)


class ImportJobModelListView(ImportJobModelViewBaseView, ListView):
    PAGE_TITLE = _('Data Import Jobs')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE
    }
    context_object_name = 'import_jobs'
    template_name = 'django_ledger/data_import/data_import_job_list.html'


class ImportJobModelUpdateView(ImportJobModelViewBaseView, UpdateView):
    template_name = 'django_ledger/data_import/import_job_update.html'
    context_object_name = 'import_job_model'
    pk_url_kwarg = 'job_pk'
    form_class = ImportJobModelUpdateForm

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['page_title'] = 'Import Job Update'
        ctx['header_title'] = 'Import Job Update'
        ctx['header_subtitle'] = self.object.description
        ctx['header_subtitle_icon'] = 'solar:import-bold'
        return ctx

    def get_success_url(self):
        entity_model = self.get_authorized_entity_instance()
        return reverse(
            viewname='django_ledger:data-import-jobs-list',
            kwargs={
                'entity_slug': entity_model.slug
            }
        )

    def form_valid(self, form):
        messages.add_message(
            self.request,
            level=messages.SUCCESS,
            message=_(f'Successfully updated Import Job {self.object.description}'),
            extra_tags='is-success'
        )
        return super().form_valid(form=form)


class ImportJobModelDeleteView(ImportJobModelViewBaseView, DeleteView):
    template_name = 'django_ledger/data_import/import_job_delete.html'
    context_object_name = 'import_job_model'
    pk_url_kwarg = 'job_pk'

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['page_title'] = 'Delete Import Job'
        ctx['header_title'] = 'Delete Import Job'
        ctx['header_subtitle'] = self.object.description
        ctx['header_subtitle_icon'] = 'solar:import-bold'
        return ctx

    def get_success_url(self):
        return reverse(
            viewname='django_ledger:data-import-jobs-list',
            kwargs={
                'entity_slug': self.AUTHORIZED_ENTITY_MODEL.slug
            }
        )


class DataImportJobDetailView(ImportJobModelViewBaseView, DetailView):
    template_name = 'django_ledger/data_import/data_import_job_txs.html'
    PAGE_TITLE = _('Import Job Staged Txs')
    context_object_name = 'import_job'
    pk_url_kwarg = 'job_pk'
    import_transactions = False
    form_class = StagedTransactionModelFormSet
    http_method_names = ['get', 'post']

    def get_form_kwargs(self):
        return {
            'entity_model': self.get_authorized_entity_instance(),
            'import_job_model': self.get_object(),
        }

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        # job_model: ImportJobModel = getattr(self, 'object', self.get_object())
        job_model: ImportJobModel = self.object

        context['page_title'] = job_model.description
        context['header_title'] = self.PAGE_TITLE
        context['header_subtitle'] = job_model.description
        context['header_subtitle_icon'] = 'tabler:table-import'

        staged_txs_formset = StagedTransactionModelFormSet(
            entity_model=self.get_authorized_entity_instance(),
            import_job_model=job_model
        )

        context['staged_txs_formset'] = staged_txs_formset

        return context

    def post(self, request, **kwargs):
        self.object = self.get_object()

        txs_formset = StagedTransactionModelFormSet(
            entity_model=self.get_authorized_entity_instance(),
            import_job_model=self.object,
            data=request.POST
        )

        # if txs_formset.is_valid():
        for tx_form in txs_formset:
            if tx_form.has_changed():
                # perform work only if form has changed...
                if tx_form.is_valid():
                    tx_form.save()
                    # import entry was selected to be split....
                    is_split = tx_form.cleaned_data['tx_split'] is True
                    if is_split:
                        tx_form.instance.add_split()

                    # import entry was selected for import...
                    is_import = tx_form.cleaned_data['tx_import']
                    if is_import:
                        # all entries in split will be going so the same journal entry... (same unit...)
                        is_bundled = tx_form.cleaned_data['bundle_split']
                        tx_form.instance.migrate() if is_bundled else tx_form.instance.migrate(split_txs=True)

        messages.add_message(request,
                             messages.SUCCESS,
                             'Successfully saved transactions.',
                             extra_tags='is-success')

        context = self.get_context_data(**kwargs)
        return self.render_to_response(context)
